import spekk.transformations
from spekk.transformations import *

from vbeam.fastmath import numpy as np


def do_nothing(f):
    return f


class ForAll(spekk.transformations.ForAll):
    def __post_init__(self):
        self.vmap_impl = np.vmap
        super().__post_init__()


class Reduce(spekk.transformations.Reduce):
    def __post_init__(self):
        self.reduce_impl = np.reduce
        super().__post_init__()


__all__ = [*spekk.transformations.__all__, "do_nothing"]



# Code is copied (and slightly modified) from Trax: https://github.com/google/trax/blob/78d2b6ef9070f1f267eb4a74d2ca6618c84abbb5/trax/tf_numpy/extensions/extensions.py#L1907
# Trax is licensed under Apache License 2.0: https://github.com/google/trax/blob/master/LICENSE

import sys

import tensorflow as tf

# Same as `import tensorflow.experimental.numpy as tnp`, but vscode doesn't recognize that for some reason.
import tensorflow._api.v2.experimental.numpy as tnp

# Enable NumPy behavior on Tensors.
tnp.experimental_enable_numpy_behavior()


def reraise(tp, value, tb=None):
    # Code is copied from Six: https://github.com/benjaminp/six/blob/3b7efbcca41857da03fb01f004ccc425ab82dfbf/six.py#L713
    # Siz is licensed under MIT License: https://github.com/benjaminp/six/blob/master/LICENSE
    try:
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value
    finally:
        value = None
        tb = None


def _tree_broadcast(to, s):
    """Broadcasts `s` to the nested structure `to`."""
    if not isinstance(to, (list, tuple, dict)):
        if not isinstance(s, (int, type(None))):
            raise ValueError
        return s
    if isinstance(s, (int, type(None))):
        return tf.nest.map_structure(lambda x: s, to)
    if isinstance(to, (list, tuple)):
        if len(to) != len(s):
            raise ValueError
        new_s = [_tree_broadcast(x, y) for x, y in zip(to, s)]
        if isinstance(to, tuple):
            new_s = tuple(new_s)
        return new_s
    elif isinstance(to, dict):
        return {k: _tree_broadcast(to[k], s[k]) for k in to}
    else:
        raise TypeError("Unsupported type %s" % type(to))


def vmap(f, in_axes=0, out_axes=0):
    """Returns a function that maps `f` over first dimension of inputs."""
    in_axes_flat = tf.nest.flatten(in_axes)
    if not all(isinstance(l, (type(None), int)) for l in in_axes_flat):
        raise TypeError(
            "vmap in_axes must be an int, None, or (nested) container with "
            "those types as leaves, but got {}.".format(in_axes)
        )
    if all(isinstance(l, type(None)) for l in in_axes_flat):
        raise ValueError("vmap must have at least one non-None value in in_axes")

    out_axes_flat = tf.nest.flatten(out_axes)
    if not all(isinstance(l, (type(None), int)) for l in out_axes_flat):
        raise TypeError(
            "vmap out_axes must be an int, None, or (nested) container with "
            "those types as leaves, but got {}.".format(out_axes)
        )

    def _f(*args):
        flat_args = tf.nest.flatten(args)
        try:
            f_in_axes = _tree_broadcast(args, in_axes)
        except ValueError:
            reraise(
                ValueError,
                ValueError(
                    "vmap in_axes specification must be a tree prefix of the "
                    r"corresponding value, got specification %s for value tree %s"
                    % (in_axes, args)
                ),
                sys.exc_info()[2],
            )
        f_in_axes_flat = tf.nest.flatten(f_in_axes)

        def tf_f(tf_args):
            """Function passed to tf.vectorized_map call."""
            # Note that unbatched arguments are not passed to tf_f. Here we fill thos
            # arguments back before calling `f`.
            tf_flat_args = []
            j = 0
            for arg, axis in zip(flat_args, f_in_axes_flat):
                if axis is None:
                    tf_flat_args.append(arg)
                else:
                    tf_flat_args.append(tf_args[j])
                    j += 1
            unbatched_args = tf.nest.pack_sequence_as(args, tf_flat_args)
            return f(*unbatched_args)

        # Constructs arguments to pass to `tf_f`.
        # Unbatch arguments are skipped. Arguments with non-zero axis are
        # transposed.
        tf_args = []
        for arg, axis in zip(flat_args, f_in_axes_flat):
            if axis is None:
                continue
            # arg = tnp.asarray(arg)  # NOTE: This breaks when vectorizing over extension types such as WaveData
            if axis != 0:
                arg = tnp.moveaxis(arg, axis, 0)
            tf_args.append(arg)
        # TODO(agarwal): consider creating a tf.function outside of _f and reusing
        # that to avoid overheads of re-vectorizing the code when running eagerly.
        outputs = tf.vectorized_map(tf_f, tf_args)
        try:
            f_out_axes = _tree_broadcast(outputs, out_axes)
        except ValueError:
            reraise(
                ValueError,
                ValueError(
                    "vmap out_axes specification must be a tree prefix of the "
                    r"corresponding value, got specification %s for value tree %s"
                    % (out_axes, outputs)
                ),
                sys.exc_info()[2],
            )

        def map_output(x, axis):
            """Maps output of tf.vectorized_map to the final output."""
            # x = tnp.asarray(x)  # NOTE: This breaks when vectorizing over extension types such as WaveData
            if axis is None:
                # Note that `tf.vectorized_map always batches the outputs.
                # Here we unbatch it again.
                return x[0, ...]
            elif axis == 0:
                return x
            else:
                # Need to transpose the output.
                return tnp.moveaxis(x, 0, axis)

        new_outputs = [
            map_output(output, axis)
            for output, axis in zip(
                tf.nest.flatten(outputs), tf.nest.flatten(f_out_axes)
            )
        ]
        return tf.nest.pack_sequence_as(outputs, new_outputs)

    return _f



import warnings
from dataclasses import dataclass

import jax
import jax.numpy as jnp

from vbeam.fastmath.backend import Backend

_already_traceable = set()


class JaxBackend(Backend):
    @property
    def ndarray(self):
        return jnp.ndarray

    def zeros(self, shape, dtype=None):
        return jnp.zeros(shape, dtype)

    def ones(self, shape, dtype=None):
        return jnp.ones(shape, dtype)

    @property
    def pi(self):
        return jnp.pi

    def abs(self, x):
        return jnp.abs(x)

    def exp(self, x):
        return jnp.exp(x)

    def log(self, x):
        return jnp.log(x)

    def log10(self, x):
        return jnp.log10(x)

    def sin(self, x):
        return jnp.sin(x)

    def cos(self, x):
        return jnp.cos(x)

    def tan(self, x):
        return jnp.tan(x)

    def arcsin(self, x):
        return jnp.arcsin(x)

    def arccos(self, x):
        return jnp.arccos(x)

    def arctan2(self, y, x):
        return jnp.arctan2(y, x)

    def sqrt(self, x):
        return jnp.sqrt(x)

    def sign(self, x):
        return jnp.sign(x)

    def nan_to_num(self, x, nan=0.0):
        return jnp.nan_to_num(x, nan=nan)

    def min(self, a, axis=None):
        return jnp.min(a, axis=axis)

    def max(self, a, axis=None):
        return jnp.max(a, axis=axis)

    def argmin(self, a, axis=None):
        return jnp.argmin(a, axis=axis)

    def argmax(self, a, axis=None):
        return jnp.argmax(a, axis=axis)

    def minimum(self, a, b):
        return jnp.minimum(a, b)

    def maximum(self, a, b):
        return jnp.maximum(a, b)

    def sum(self, a, axis=None):
        return jnp.sum(a, axis=axis)

    def prod(self, a, axis=None):
        return jnp.prod(a, axis=axis)

    def mean(self, a, axis=None):
        return jnp.mean(a, axis=axis)

    def var(self, a, axis=None):
        return jnp.var(a, axis=axis)

    def cumsum(self, a, axis=None):
        return jnp.cumsum(a, axis=axis)

    def cross(self, a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
        return jnp.cross(a, b, axisa=axisa, axisb=axisb, axisc=axisc, axis=axis)

    def nansum(self, a, axis=None):
        return jnp.nansum(a, axis=axis)

    def histogram(self, a, bins=10, weights=None):
        return jnp.histogram(a, bins=bins, weights=weights)

    def array(self, x, dtype=None):
        return jnp.array(x, dtype=dtype)

    def transpose(self, a, axes=None):
        return jnp.transpose(a, axes=axes)

    def swapaxes(self, a, axis1, axis2):
        return jnp.swapaxes(a, axis1, axis2)

    def moveaxis(self, a, source, destination):
        return jnp.moveaxis(a, source, destination)

    def stack(self, arrays, axis=0):
        return jnp.stack(arrays, axis=axis)

    def tile(self, A, reps):
        return jnp.tile(A, reps)

    def concatenate(self, arrays, axis=0):
        return jnp.concatenate(arrays, axis=axis)

    def meshgrid(self, *xi, indexing="xy"):
        return jnp.meshgrid(*xi, indexing=indexing)

    def linspace(self, start, stop, num=50):
        return jnp.linspace(start, stop, num=num)

    def arange(self, start, stop=None, step=None):
        return jnp.arange(start, stop, step)

    def expand_dims(self, a, axis):
        return jnp.expand_dims(a, axis=axis)

    def ceil(self, x):
        return jnp.ceil(x)

    def floor(self, x):
        return jnp.floor(x)

    def modf(self, x):
        return jnp.modf(x)

    def round(self, x):
        return jnp.round(x)

    def clip(self, a, a_min, a_max):
        return jnp.clip(a, a_min, a_max)

    def where(self, condition, x, y):
        return jnp.where(condition, x, y)

    def select(self, condlist, choicelist, default=0):
        return jnp.select(condlist, choicelist, default)

    def logical_or(self, x1, x2):
        return jnp.logical_or(x1, x2)

    def logical_and(self, x1, x2):
        return jnp.logical_and(x1, x2)

    def squeeze(self, a, axis=None):
        return jnp.squeeze(a, axis=axis)

    def ravel(self, a):
        return jnp.ravel(a)

    def take(self, a, indices, axis=None):
        if self.is_ndarray(a):
            return jnp.take(a, indices, axis=axis)
        else:
            return jax.tree_util.tree_map(lambda a: jnp.take(a, indices, axis=axis), a)

    def interp(self, x, xp, fp, left=None, right=None, period=None):
        return jnp.interp(x, xp, fp, left, right, period)

    def gather(self, a, indices):
        return a[indices]

    class add:
        @staticmethod
        def at(a, indices, b):
            return a.at[indices].add(b)

    def jit(self, fun, static_argnums=None, static_argnames=None):
        return jax.jit(
            fun,
            static_argnums=static_argnums,
            static_argnames=static_argnames,
        )

    def vmap(self, fun, in_axes, out_axes=0):
        return jax.vmap(fun, in_axes=in_axes, out_axes=out_axes)

    def scan(self, f, init, xs):
        return jax.lax.scan(f, init, xs)

    def as_traceable_dataclass_obj(self, original_obj, data_fields, aux_fields):
        cls = type(original_obj)

        # Register the class of the object (if it hasn't been done before) as a
        # pytree-node.
        if cls not in _already_traceable:

            def flatten_fn(obj):
                data = [getattr(obj, child) for child in data_fields]
                aux_data = [getattr(obj, aux_field) for aux_field in aux_fields]
                return data, (cls, aux_data)

            def unflatten_fn(treedef, flattened_obj):
                cls, aux_data = treedef
                children_kwargs = dict(zip(data_fields, flattened_obj))
                aux_kwargs = dict(zip(aux_fields, aux_data))
                kwargs = {**children_kwargs, **aux_kwargs}
                return cls(**kwargs)

            try:
                # May throw ValueError if this module got reloaded and
                # _already_traceable got reset.
                jax.tree_util.register_pytree_node(cls, flatten_fn, unflatten_fn)
            except ValueError as e:
                warnings.warn(
                    f"Received ValueError when trying to register JAX pytree: {e}"
                )
            finally:
                _already_traceable.add(cls)  # Don't register it again.

        as_dataclass = dataclass(cls)  # Make it a dataclass as well.
        original_obj.__class__ = as_dataclass
        return original_obj



import os
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "vbeam"
copyright = "2023, Magnus Dalen Kvalevåg"
author = "Magnus Dalen Kvalevåg"

# The full version, including alpha/beta/rc tags
release = "1.0.5"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # Generates documentation from docstrings
    "sphinx.ext.napoleon",  # Adds support for NumPy and Google style docstrings
    "sphinx.ext.autosummary",  # Automatically generates documentation for modules
    "sphinx.ext.viewcode",
    "sphinx_copybutton",  # Adds copy-button to code-blocks
    "myst_nb",  # Rendering Jupyter notebooks

    # Only works on read-the-docs
    # "sphinxcontrib.jquery",  # Adds jQuery (needed by hoverxref)
    # "hoverxref.extension",  # Hover over references
]

autodo_mock_imports = ["vbeam.core"]
autodoc_member_order = "bysource"
# autosummary_generate = True
autosummary_ignore_module_all = False

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_book_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_logo = "_static/vbeam_header.png"
html_favicon = "_static/favicon-92x92.png"


hoverxref_roles = [
    "term",
]
hoverxref_role_types = {
    "term": "tooltip",
}


from typing import Tuple

from vbeam.core import InterpolationSpace1D
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util import ensure_positive_index


@traceable_dataclass(data_fields=("min", "d", "n"))
class FastInterpLinspace(InterpolationSpace1D):
    """Interpolation for linspace.

    If we assume that it is uniformly spaced (linspace), an array can be interpolated
    really quickly by subtracting the minimum value and dividing by the step-size."""

    min: float
    d: float
    n: int

    @staticmethod
    def from_array(arr: np.ndarray) -> "FastInterpLinspace":
        # Assumes that arr was created as a linspace.
        return FastInterpLinspace(arr[0], arr[1] - arr[0], len(arr))

    def to_array(self) -> np.ndarray:
        return np.linspace(self.min, self.min + self.d * self.n, self.n)

    def interp1d_indices(self, x: np.ndarray) -> Tuple[float, int, int, float, float]:
        """Return a tuple of 5 elements with information about how to interpolate the
        array.

        Returns a tuple of:
          0. bounds_flag: If 0, then x is within bounds. If -1, then x is outside of
               bounds on the left side. If 1, then x is out of bounds on the right side.
          1. clipped_i1: The index of the first value used in the interpolation.
          2. clipped_i2: The index of the second value used in the interpolation.
          3. p1: The weighting of the value at index clipped_i1.
          4. p2: The weighting of the value at index clipped_i2.

        The final interpolated value can be found as such:
        >>> fp = np.array([0, 1, 2])
        >>> interp = FastInterpLinspace.from_array(fp)
        >>> x = np.array([0.5, 1.5])
        >>> bounds_flag, clipped_i1, clipped_i2, p1, p2 = interp.interp1d_indices(x)
        >>> interpolated_values = fp[clipped_i1]*p1 + fp[clipped_i2]*p2
        >>> interpolated_values[1]
        1.5
        """
        pseudo_index = (x - self.min) / self.d
        i_floor = np.floor(pseudo_index)
        di = pseudo_index - i_floor

        bounds_flag = 0
        bounds_flag = np.where(pseudo_index < 0, -1, bounds_flag)
        bounds_flag = np.where(pseudo_index > (self.n - 1), 1, bounds_flag)
        clipped_i1 = np.clip(i_floor, 0, self.n - 1).astype("int32")
        clipped_i2 = np.clip(i_floor + 1, 0, self.n - 1).astype("int32")
        p1, p2 = (1 - di), di
        return bounds_flag, clipped_i1, clipped_i2, p1, p2

    def interp1d(
        self,
        x: np.ndarray,
        fp: np.ndarray,
        left: int = 0,
        right: int = 0,
    ) -> np.ndarray:
        bounds_flag, clipped_i1, clipped_i2, p1, p2 = self.interp1d_indices(x)
        bounds_flag = np.expand_dims(bounds_flag, tuple(range(1, fp.ndim)))
        p1 = np.expand_dims(p1, tuple(range(1, fp.ndim)))
        p2 = np.expand_dims(p2, tuple(range(1, fp.ndim)))
        v = fp[clipped_i1] * p1 + fp[clipped_i2] * p2
        v = np.where(bounds_flag == -1, left, v)
        v = np.where(bounds_flag == 1, right, v)
        return v

    # InterpolationSpace1D interface
    def __call__(self, x, fp):
        return self.interp1d(x, fp)

    @property
    def start(self) -> float:
        return self.min

    @property
    def end(self) -> float:
        return (self.d * self.n) + self.min

    @staticmethod
    def interp2d(
        x: np.ndarray,
        y: np.ndarray,
        xp: "FastInterpLinspace",
        yp: "FastInterpLinspace",
        z: np.ndarray,
        azimuth_axis: int = 0,
        depth_axis: int = 1,
        *,  # Remaining args must be passed by name (to avoid confusion)
        padding: float = 0.0,
    ) -> np.ndarray:
        # Ensure that the axes are positive numbers
        azimuth_axis = ensure_positive_index(z.ndim, azimuth_axis)
        depth_axis = ensure_positive_index(z.ndim, depth_axis)

        # Ensure that the azimuth and depth axes are the first two axes
        if depth_axis == 0 and azimuth_axis == 1:
            z = np.swapaxes(z, azimuth_axis, depth_axis)
        else:
            z = np.moveaxis(z, azimuth_axis, 0)
            z = np.moveaxis(z, depth_axis, 1)

        # Interpolate along the axes
        bounds_flag_x, clipped_xi1, clipped_xi2, px1, px2 = xp.interp1d_indices(x)
        bounds_flag_y, clipped_yi1, clipped_yi2, py1, py2 = yp.interp1d_indices(y)

        # Make weighting values and flags broadcastable with respect to the shape of
        # elements in z. This makes is it so that we can interpolate grids of vectors;
        # not just grids of scalars.
        num_value_dims = len(z.shape[2:])
        broadcastable = lambda a: np.expand_dims(
            a, tuple(range(a.ndim, a.ndim + num_value_dims))
        )
        px1, px2, py1, py2 = [broadcastable(p) for p in [px1, px2, py1, py2]]
        bounds_flag_x = broadcastable(bounds_flag_x)
        bounds_flag_y = broadcastable(bounds_flag_y)

        v0 = z[clipped_xi1, clipped_yi1] * px1 + z[clipped_xi2, clipped_yi1] * px2
        v1 = z[clipped_xi1, clipped_yi2] * px1 + z[clipped_xi2, clipped_yi2] * px2
        v = v0 * py1 + v1 * py2
        v = np.where(np.logical_or(bounds_flag_x != 0, bounds_flag_y != 0), padding, v)

        # Swap axes back to their original positions
        if depth_axis == 0 and azimuth_axis == 1:
            v = np.swapaxes(v, azimuth_axis, depth_axis)
        elif v.ndim >= 2:
            v = np.moveaxis(v, 1, depth_axis)
            v = np.moveaxis(v, 0, azimuth_axis)
        return v

    @property
    def shape(self):
        return self.min.shape

    @property
    def ndim(self):
        return len(self.shape)



from typing import Optional, Sequence

from vbeam.fastmath import numpy as np


def grid(*axes: np.ndarray, shape: Optional[Sequence[int]] = None) -> np.ndarray:
    """Return an array of each point position, organized in a grid.

    >>> x, z = np.array([1,2,3]), np.array([1,2,3,4])
    >>> points = grid(x, z)
    >>> points.shape == (3, 4, 2)
    True
    >>> points.shape == (x.size, z.size, 2)
    True
    >>> (points[:, 0, 0] == x).all()
    True
    >>> (points[0, :, 1] == z).all()
    True

    We can control the final shape of the grid by specifying the shape argument:
    >>> points_3d = grid(x, np.array([0.0]), z, shape=(3, 4, 3))
    >>> (points_3d[:, :, [0, 2]] == points).all()
    True
    """
    points = np.stack(np.meshgrid(*axes, indexing="ij"), axis=-1)
    if shape:
        points = points.reshape(shape)
    return points



import os
import sys

# Disable GPU for tests
os.environ["CUDA_VISIBLE_DEVICES"] = ""

# Add helper functions to path
sys.path.append(os.path.join(os.path.dirname(__file__), "vbeam_test_helpers"))

import pytest

from vbeam.fastmath import backend_manager
from vbeam.fastmath import numpy as global_np_backend

# backends = ["numpy", "jax", "tensorflow"]
backends = ["numpy", "jax"]


@pytest.fixture(params=backends)
def np(request):
    """A fixture that provides a fastmath backend to be used.

    Usage:
    def my_test(np: Backend):  # np has type vbeam.fastmath.Backend
        assert np.abs(-1) == 1

    tests/conftest.py (this file) is automatically imported by pytest when running tests
    so this function will be available to all tests automatically.
    """
    with backend_manager.using_backend(request.param):
        yield global_np_backend


@pytest.fixture(params=("jit", "no_jit"))
def jit_able(request, np):
    def _jit(f):
        if request.param == "no_jit":
            return f
        elif backend_manager._active_backend == "jax":
            import jax

            return jax.jit(f)
        elif backend_manager._active_backend == "tensorflow":
            import tensorflow as tf

            return tf.function(f, jit_compile=True)
        else:
            return f  # Do nothing

    return _jit



from collections import namedtuple
from functools import partial

import jax
import jax.numpy as jnp

from vbeam.fastmath import backend_manager
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.geometry import Line

backend_manager.active_backend = "jax"

# TODO: Make vbeam.util.geometry.Line work with 2D points
@traceable_dataclass(("a", "b", "c"))
class LineNB:
    a: float
    b: float
    c: float

    @staticmethod
    def passing_through(point1: np.ndarray, point2: np.ndarray) -> "Line":
        "Construct a line that passes through both point1 and point2."
        x1, z1 = point1
        x2, z2 = point2
        a = z1 - z2
        b = x2 - x1
        c = x1 * z2 - x2 * z1
        return Line(a, b, c)

    @staticmethod
    def from_anchor_and_angle(anchor: np.ndarray, angle: float) -> "Line":
        "Construct a line that passes through the anchor and has the given angle."
        return Line.passing_through(
            anchor, anchor + np.array([np.cos(angle), np.sin(angle)])
        )

    def intersection(l1: "Line", l2: "Line") -> np.ndarray:
        "Return the point where the lines l1 and l2 intersect."
        x = (l1.b * l2.c - l2.b * l1.c) / (l1.a * l2.b - l2.a * l1.b)
        z = (l2.a * l1.c - l1.a * l2.c) / (l1.a * l2.b - l2.a * l1.b)
        return np.array([x, z])

    @property
    def angle(self):
        "The angle of the line."
        return np.arctan2(self.b, self.a)

    def signed_distance(self, point: np.ndarray) -> float:
        """Return the signed distance between the point and the nearest point on the
        line.

        The sign of the distance will be positive on the left-hand side and negative on
        the right-hand side of the line.

        The y-component of the point is ignored.

        >>> vertical_line = Line.passing_through(np.array([0,0,0]), np.array([0,0,1]))
        >>> vertical_line.signed_distance(np.array([0,0,5]))
        0.0
        >>> vertical_line.signed_distance(np.array([-1,0,5]))  # The left-hand side
        1.0
        >>> vertical_line.signed_distance(np.array([1,0,5]))  # The right-hand side
        -1.0
        """
        norm = np.sqrt(self.a**2 + self.b**2)
        return (self.a * point[..., 0] + self.b * point[..., 1] + self.c) / norm


Vertex = namedtuple("Vertex", ["xyz", "value"])
Polygon = namedtuple("Polygon", ["v1", "v2", "v3", "v4"])


@jax.jit
def image2polygons(coords, image) -> Polygon:
    @partial(jax.vmap, in_axes=[0, 0])
    def polygon_from_coord(xi: int, zi: int) -> Polygon:
        return Polygon(
            *(
                Vertex(coords[xi + xd, zi + zd], image[xi + xd, zi + zd])
                for (xd, zd) in [(0, 0), (1, 0), (1, 1), (0, 1)]
            )
        )

    nx, nz = image.shape
    X, Z = jnp.meshgrid(jnp.arange(nx - 1), jnp.arange(nz - 1), indexing="ij")
    return polygon_from_coord(jnp.ravel(X), jnp.ravel(Z))


def contains(polygon: Polygon, point: jnp.ndarray):
    "Return True if the point is within the polygon."
    l1 = Line.passing_through(polygon.v1.xyz, polygon.v2.xyz)
    l2 = Line.passing_through(polygon.v2.xyz, polygon.v3.xyz)
    l3 = Line.passing_through(polygon.v3.xyz, polygon.v4.xyz)
    l4 = Line.passing_through(polygon.v4.xyz, polygon.v1.xyz)
    return (
        (l1.signed_distance(point) >= 0)
        & (l2.signed_distance(point) >= 0)
        & (l3.signed_distance(point) >= 0)
        & (l4.signed_distance(point) >= 0)
    )


def lerp_triangle(point, v1: Vertex, v2: Vertex, v3: Vertex):
    # https://codeplea.com/triangular-interpolation
    denominator = (v2.xyz[2] - v3.xyz[2]) * (v1.xyz[0] - v3.xyz[0]) + (
        v3.xyz[0] - v2.xyz[0]
    ) * (v1.xyz[2] - v3.xyz[2])
    w1 = (
        (v2.xyz[2] - v3.xyz[2]) * (point[0] - v3.xyz[0])
        + (v3.xyz[0] - v2.xyz[0]) * (point[2] - v3.xyz[2])
    ) / denominator
    w2 = (
        (v3.xyz[2] - v1.xyz[2]) * (point[0] - v3.xyz[0])
        + (v1.xyz[0] - v3.xyz[0]) * (point[2] - v3.xyz[2])
    ) / denominator
    w3 = 1 - w1 - w2
    value = w1 * v1.value + w2 * v2.value + w3 * v3.value
    return jnp.where((w1 >= 0) & (w2 >= 0) & (w3 >= 0), value, 0)


def lerp(poly: Polygon, point: jnp.ndarray):
    return lerp_triangle(point, poly.v1, poly.v2, poly.v3) + lerp_triangle(
        point, poly.v1, poly.v3, poly.v4
    )


@jax.jit
@partial(jax.vmap, in_axes=(None, 0))
def lerp_all(polygons: Polygon, points: jnp.ndarray):
    vmapped_lerp = jax.vmap(lerp, (0, None))
    return jnp.sum(vmapped_lerp(polygons, points))



from typing import Sequence, TypeVar, Union

TSelf = TypeVar("TSelf", bound="KernelData")


class KernelData:
    """KernelData is the base class for containers of data that is needed by a kernel.

    KernelData provides methods that lets us treat it like a mapping, meaning that we
    can easily pass the whole kernel data object as kwargs to a kernel function,
    e.g.: signal_for_point(**kernel_data)."""

    # Defining keys and __getitem__ means this can be used as a mapping
    def keys(self) -> Sequence[str]:
        raise NotImplementedError

    def __getitem__(self, key: Union[str, int]):
        if isinstance(key, int):
            key = list(self.keys())[key]
        return getattr(self, key)

    def copy(self: TSelf) -> TSelf:
        raise NotImplementedError



from vbeam.core import Apodization
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass()
class NoApodization(Apodization):
    """Always return 1.0 (no weighting)."""

    def __call__(self, *_, **__) -> float:
        return 1.0



import pytest
from numpy import allclose, random
from vbeam.fastmath import backend_manager
from vbeam.fastmath import numpy as np


def test_concatenate_general():
    import tensorflow as tf

    # Test some randomly generated cases
    def assert_equal_results_numpy_tf(arr):
        # Test concatenating along each possible axis
        for axis in range(arr.ndim - 1):
            with backend_manager.using_backend("numpy"):
                expected = np.concatenate(arr, axis)

            with backend_manager.using_backend("tensorflow"):
                assert allclose(expected, np.concatenate(arr, axis))
                jitted_concatenate = tf.function(np.concatenate, jit_compile=True)
                assert allclose(expected, jitted_concatenate(arr, axis))

    assert_equal_results_numpy_tf(random.random((6, 7, 8, 9, 10)))
    assert_equal_results_numpy_tf(random.random((6, 1)))
    assert_equal_results_numpy_tf(random.random((6, 0)))


def test_concatenate_negative_axes():
    arrays = random.random((6, 7, 8, 9, 10))
    with backend_manager.using_backend("numpy"):
        expected0 = np.concatenate(arrays, -1)
        expected1 = np.concatenate(arrays, -2)
        expected2 = np.concatenate(arrays, -3)
        expected3 = np.concatenate(arrays, -4)
    with backend_manager.using_backend("tensorflow"):
        assert allclose(expected0, np.concatenate(arrays, -1))
        assert allclose(expected1, np.concatenate(arrays, -2))
        assert allclose(expected2, np.concatenate(arrays, -3))
        assert allclose(expected3, np.concatenate(arrays, -4))


def test_concatenate_raising_errors():
    # Test raising errors
    with backend_manager.using_backend("tensorflow"):
        with pytest.raises(
            ValueError, match="zero-dimensional arrays cannot be concatenated"
        ):
            np.concatenate(np.ones((2,)))
        with pytest.raises(
            ValueError, match="axis 1 is out of bounds for array of dimension 1"
        ):
            np.concatenate(np.ones((2, 2)), 1)
        with pytest.raises(
            ValueError, match="axis -2 is out of bounds for array of dimension 1"
        ):
            np.concatenate(np.ones((2, 2)), -2)



import numpy
from spekk import trees


def assert_equal_treedefs(a: trees.Tree, b: trees.Tree):
    if trees.has_treedef(a):
        assert trees.has_treedef(b)
        a = trees.treedef(a)
        b = trees.treedef(b)
        assert a.keys() == b.keys()
        for k in a.keys():
            assert_equal_treedefs(a.get(k), b.get(k))
    else:
        numpy.testing.assert_equal(a, b)



from hypothesis import strategies as st

from vbeam.fastmath import numpy as np
from vbeam.scan import sector_scan


def linspaces(min: float, max: float, num: int = 50, ensure_increasing: bool = False):
    "Generate linspaces where start <= stop."
    return (
        st.tuples(st.floats(min, max), st.floats(min, max))
        .filter(lambda args: not ensure_increasing or args[0] <= args[1])
        .map(lambda args: np.linspace(*args, num))
    )


def sector_scans(shape):
    "Generate instances of :class:`~vbeam.scan.SectorScan` with the given shape."
    assert len(shape) in (2, 3), "scan must be 2D or 3D"
    return st.builds(
        sector_scan,
        *[linspaces(-np.pi, np.pi, size) for size in shape[:-1]],
        linspaces(0, 2, shape[-1]),
        apex=st.builds(
            lambda x, z: np.array([x, 0.0, z]),
            st.floats(-4, 4),
            st.floats(-4, 4),
        ),
    )



from typing import Callable, Tuple

from vbeam.fastmath import numpy as np


def get_window_indices(
    image_shape: Tuple[int, ...],
    window_radius: Tuple[int, ...],
) -> np.ndarray:
    indices = np.stack(
        np.meshgrid(*(np.arange(0, s) for s in image_shape), indexing="ij"), -1
    )
    # Make it broadcastable
    indices = indices[(..., *(None for _ in range(len(window_radius))), slice(None))]
    indices_window_offset = np.stack(
        np.meshgrid(*(np.arange(-r, r + 1) for r in window_radius), indexing="ij"), -1
    )
    return indices + indices_window_offset


def windowed(f: Callable, window_radius: Tuple[int, ...]) -> np.ndarray:
    def apply_windowed(image, *args, **kwargs):
        indices = get_window_indices(image.shape, window_radius)

        def apply1(indices):
            windowed_image = image[indices[..., 0], indices[..., 1]]
            return f(windowed_image, *args, **kwargs)

        # vmap over all axes in the image
        for _ in range(image.ndim):
            apply1 = np.vmap(apply1, [0])
        return apply1(indices)

    return apply_windowed



"""Interfaces and kernel function for the core of the beamforming algorithm.

``vbeam`` is built around a small functional core found in this package. It is built 
around the :func:`~vbeam.core.kernels.signal_for_point` kernel [#]_ which depend on a 
set of interfaces to calculate the delayed, interpolated, and weighted signal for a 
given point.

A generalized delay-and-sum beamformer can be expressed as in equation 
:eq:`das_general_beamformer`:

.. math::
    :label: das_general_beamformer

    b_{DAS}(p) = 
    \\sum_{a=0}^{N-1} 
        \\sum_{m=0}^{M-1}
            w(a, m, p)\\,
            s_{a, m}(t)\\,
            e^{i2\\pi f_{mod}t},
    t := \\frac{d(a, m, p)}{c}

where :math:`a`, :math:`m`, and :math:`p` represent a given :term:`transmit`, 
:term:`receiver`, and :term:`point`, respectively. Note that this equation is for a 
full delay-and-sum (DAS) beamformer while the 
:func:`~vbeam.core.kernels.signal_for_point` function is only the inner part of the 
loop, i.e.: it computes the delayed signal for a single transmit, receiver, and point. 
For building up a full DAS beamformer from 
:func:`~vbeam.core.kernels.signal_for_point`, see :mod:`vbeam.beamformers`.

In equation :eq:`das_general_beamformer`:

* :math:`w(a, m, p)` weights the delayed and interpolated signal and is represented by 
  the :class:`~vbeam.core.apodization.Apodization` class in vbeam.

* :math:`s_{a, m}(t)` is the delayed and interpolated :term:`signal` for the 
  :term:`transmit` and :term:`receiver` at time :math:`t` seconds *(defined below)*. 
  Interpolation is represented by the 
  :class:`~vbeam.core.interpolation.InterpolationSpace1D` class.

* :math:`e^{i2\\pi f_{mod}t}` is the modulation frequency correction, which should be 
  applied if the signal is a demodulated IQ signal. :math:`f_{mod}` is the modulation 
  frequency, represented as a number, and given as an argument to 
  :func:`~vbeam.core.kernels.signal_for_point`.

* :math:`d(a, m, p)` is the distance the wave has travelled from the :term:`sender` to 
  the :term:`point`, and back to the :term:`receiver`. This is represented by the 
  :class:`~vbeam.core.wavefront.ReflectedWavefront` and 
  :class:`~vbeam.core.wavefront.TransmittedWavefront` classes.

* :math:`t := \\frac{d(a, m, p)}{c}` is the delay in seconds, computed by dividing the 
  distance by the speed of sound. The speed of sound may be just a number, or in  more 
  advanced cases, an instance of the :class:`~vbeam.core.speed_of_sound.SpeedOfSound` 
  class.

Again, it is important to keep in mind that 
:func:`~vbeam.core.kernels.signal_for_point` runs for just a single :term:`point`, a 
single :term:`receiving element<Receiver>`, a single 
:term:`transmitted wave<Transmit>`, etc. In general, it returns a scalar value.

.. rubric:: Footnotes
.. [#] We refer to :func:`~vbeam.core.kernels.signal_for_point` as a 
       :term:`kernel<Kernel function>` because it is the central function that we 
       repeatedly call when beamforming. However, it is just a normal Python function — 
       it just happens to be a very important function.
"""

from vbeam.core.apodization import Apodization
from vbeam.core.element_geometry import ElementGeometry
from vbeam.core.interpolation import InterpolationSpace1D
from vbeam.core.kernels import KernelData, SignalForPointData, signal_for_point
from vbeam.core.speed_of_sound import SpeedOfSound
from vbeam.core.wave_data import WaveData
from vbeam.core.wavefront import (
    MultipleTransmitDistances,
    ReflectedWavefront,
    TransmittedWavefront,
)

__all__ = [
    "Apodization",
    "ElementGeometry",
    "InterpolationSpace1D",
    "KernelData",
    "SignalForPointData",
    "signal_for_point",
    "SpeedOfSound",
    "WaveData",
    "MultipleTransmitDistances",
    "ReflectedWavefront",
    "TransmittedWavefront",
]



import os
import time
from pathlib import PurePosixPath
from typing import Sequence, Union
from urllib.parse import urlparse, urlsplit
from urllib.request import urlretrieve


class DownloadReporter:
    """Class for printing the progress of downloading a file."""

    def __init__(self, filepath: str):
        self.filepath = filepath
        self.throttle_s = 0.1
        self.prev_progress = time.time() - self.throttle_s
        self.show_progress(0)

    def show_progress(self, percentage, final=False):
        if (time.time() - self.prev_progress) >= self.throttle_s or final:
            print(
                f'Downloading file to "{self.filepath}": {percentage:.1f}%',
                end="\r" if not final else None,
            )
            self.prev_progress = time.time()

    def __call__(self, block_num: int, read_size: int, total_file_size: int):
        percentage = min(block_num * read_size, total_file_size) / total_file_size * 100
        self.show_progress(percentage, final=block_num * read_size >= total_file_size)


def cached_download(
    url: str,
    local_dir: str = None,
    filename: Union[str, Sequence[str]] = None,
    invalidate_cache: bool = False,
) -> str:
    """
    Download and cache the file at URL, if not already cached. Return the path to the 
    file.

    Args:
      url: The URL of the file to be downloaded.
      local_dir: Optional base-directory of where to store the file. If not provided,
        either the environment variable VBEAM_DOWNLOADS or "~/.vbeam_downloads" is used.
      filename: Optional name of the downloaded file. Can be a string, or a sequence of
        strings representing the full path, e.g.:
          ("foo", "bar", "baz.txt") -> "foo/bar/baz.txt".
        If not provided, the directories and filename are inferred from the URL, e.g.:
          "https://www.example.com/foo/bar/baz.txt" -> "www.example.com/foo/bar/baz.txt"
      invalidate_cache: If True, forces (re)downloading the file from the URL.

    Returns:
      The path to the cached file downloaded from the URL.
    """
    # Read from VBEAM_DOWNLOADS environment variable if local_dir is None, or set it to
    # the default value of "~/.vbeam_downloads".
    if not local_dir:
        local_dir = os.environ.get("VBEAM_DOWNLOADS", None) or "~/.vbeam_downloads"

    local_dir = os.path.expanduser(local_dir)  # Expand tilde (~) in path
    if not os.path.isdir(local_dir):
        os.makedirs(local_dir, exist_ok=True)

    # Convert filename to a sequence of directory names, followed by the name of the
    # file, if filename is None or a string.
    if filename is None:
        parsed_url = urlparse(url)
        filename = (parsed_url.hostname,) + PurePosixPath(urlsplit(url).path).parts[1:]
    elif isinstance(filename, str):
        filename = (filename,)

    filepath = f"{local_dir}/{'/'.join(filename)}"
    filedir = f"{local_dir}/{'/'.join(filename[:-1])}"
    if not os.path.exists(filepath) or invalidate_cache:
        os.makedirs(filedir, exist_ok=True)
        urlretrieve(url, filepath, reporthook=DownloadReporter(filepath))

    return filepath



from typing import Optional

from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass(("transmit", "receive"))
class TxRxApodization(Apodization):
    """Apodization for both transmit and receive (just a container of two Apodization
    functions)."""

    transmit: Optional[Apodization] = None
    receive: Optional[Apodization] = None

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        weight = 1.0
        if self.transmit:
            weight *= self.transmit(sender, point_position, receiver, wave_data)
        if self.receive:
            weight *= self.receive(sender, point_position, receiver, wave_data)
        return weight



"""This module implements some popular window functions (also called apodization 
functions or tapering functions).

See the notebook ``docs/tutorials/apodization/windows.ipynb`` for a visualization of 
the various implemented windows."""

from abc import ABC, abstractmethod

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


class Window(ABC):
    """A window (also called apodization function or tapering function) is used to get
    more desirable main-lobe/side-lobe characteristics. A Window object can be called
    as afunction. It takes a number between 0 and 0.5 and returns the weight of the
    window, where the function is highest at 0 and lowest at 0.5.

    For example, the output from a Bartlett window may look like this:
    >>> window = Bartlett()
    >>> ratios = np.linspace(0, 0.5, 6)
    >>> weights = [window(r) for r in ratios]
    >>> [f"{w:.1f}" for w in weights]  # Round to 1 decimal to avoid numerical errors
    ['1.0', '0.8', '0.6', '0.4', '0.2', '0.0']
    """

    @abstractmethod
    def __call__(self, ratio: float) -> float:
        """Return the weight for the ratio (between 0 and 0.5). The peak is at ratio=0,
        and it tapers off as the ratio approaches 0.5."""


@traceable_dataclass()
class NoWindow(Window):
    def __call__(self, ratio: float) -> float:
        return np.ones(ratio.shape)


def _within_valid(ratio: float) -> bool:
    return np.logical_and(0 <= ratio, ratio <= 0.5)


@traceable_dataclass()
class Rectangular(Window):
    def __call__(self, ratio: float) -> float:
        return _within_valid(ratio) * 1.0


@traceable_dataclass(("a0", "a1"))
class Hanning(Window):
    a0: float = 0.5
    a1: float = 0.5

    def __call__(self, ratio: float) -> float:
        return _within_valid(ratio) * (self.a0 + self.a1 * np.cos(2 * np.pi * ratio))


@traceable_dataclass()
class Hamming(Window):
    def __call__(self, ratio: float) -> float:
        return Hanning(0.53836, 0.46164)(ratio)


@traceable_dataclass(("roll",))
class Tukey(Window):
    roll: float

    def __call__(self, ratio: float) -> float:
        p1 = ratio <= (1 / 2 * (1 - self.roll))
        p2 = ratio > (1 / 2 * (1 - self.roll))
        p3 = (ratio < (1 / 2)) * 0.5
        p4 = 1 + np.cos(2 * np.pi / self.roll * (ratio - self.roll / 2 - 1 / 2))
        return _within_valid(ratio) * (p1 + p2 * p3 * p4)


def Tukey25() -> Tukey:
    return Tukey(0.25)


def Tukey50() -> Tukey:
    return Tukey(0.5)


def Tukey75() -> Tukey:
    return Tukey(0.75)


def Tukey80() -> Tukey:
    return Tukey(0.8)


@traceable_dataclass()
class Bartlett(Window):
    def __call__(self, ratio: float) -> float:
        return _within_valid(ratio) * (0.5 - ratio) * 2


if __name__ == "__main__":
    import doctest

    from vbeam.fastmath import backend_manager

    with backend_manager.using_backend("numpy"):
        doctest.testmod()



import math
from typing import Optional, Sequence

from spekk import Spec

from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.util.transformations import *


def get_apodization_values(
    apodization: Apodization,
    sender: ElementGeometry,
    point_position: np.ndarray,
    receiver: ElementGeometry,
    wave_data: WaveData,
    spec: Spec,
    dimensions: Optional[Sequence[str]] = None,
    average: bool = False,
    jit: bool = True,
):
    """
    Calculate and return the apodization values based on the provided arguments
    (``sender``, ``point_position``, ``receiver``, and ``wave_data``).

    The ``dimensions`` argument determines what dimensions to keep; all others are
    summed over (except when ``dimesions`` is None where we keep all dimensions
    instead). ``spec`` describes the dimensions of the data. E.g., if ``dimensions`` is
    ``["transmits", "x", "z"]``, the result will be a 3D array with shape (Nt, Nx, Nz),
    where Nt, Nx, Nz are the sizes of the given dimensions in the data.

    Args:
        apodization (Apodization): The apodization function to use.
        sender (ElementGeometry): The sender argument to ``apodization``.
        point_position (np.ndarray): The point_position argument to ``apodization``.
        receiver (ElementGeometry): The receiver argument to ``apodization``.
        wave_data (WaveData): The wave data argument to ``apodization``.
        spec (Spec): A spec describing the dimensions/shape of the arguments.
        dimensions (Optional[Sequence[str]]): The dimensions to keep in the returned
            result. If it is an empty list, all dimensions are summed over. If it is
            None, all dimensions from the spec are kept.
        average (bool): If True, the result is averaged instead of summed.
        jit (bool): If True, the process is JIT-compiled (if the backend supports it).

    Returns:
        np.ndarray: The calculated apodization values with shape corresponding to the
        dimensions defined in ``dimensions``.
    """
    kwargs = {
        "apodization": apodization,
        "sender": sender,
        "point_position": point_position,
        "receiver": receiver,
        "wave_data": wave_data,
    }

    # Return the full datacube if dimensions are not given
    # Careful! This may allocate a lot of memory.
    if dimensions is None:
        dimensions = list(spec.dimensions)

    # Define what dimensions to vmap and sum over and how
    vmap_dimensions = (
        spec["apodization"].dimensions
        | spec["sender"].dimensions
        | spec["point_position"].dimensions
        | spec["receiver"].dimensions
        | spec["wave_data"].dimensions
    )
    sum_dimensions = vmap_dimensions - set(dimensions)
    reduce_sum_dimension = spec["wave_data"].dimensions & sum_dimensions
    if reduce_sum_dimension:
        # Make it only one of the dimensions (doesn't matter which one)
        reduce_sum_dimension = {reduce_sum_dimension.pop()}
        vmap_dimensions -= reduce_sum_dimension

    # Define how to calculate the apodization values
    calculate_apodization = compose(
        lambda apodization, *args, **kwargs: apodization(*args, **kwargs),
        *[ForAll(dim) for dim in vmap_dimensions],
        Apply(np.sum, [Axis(dim) for dim in sum_dimensions - reduce_sum_dimension]),
        # [*reduce_sum_dimension][0] gets the "first element" of the set
        Reduce.Sum([*reduce_sum_dimension][0]) if reduce_sum_dimension else do_nothing,
        # Put the dimensions in the order defined by keep
        Apply(np.transpose, [Axis(dim, keep=True) for dim in dimensions]),
        # Make it run faster if `jit` is True and if the backend supports it.
        Wrap(np.jit) if jit else do_nothing,
    ).build(spec)

    # Calculate the apodization values
    values = calculate_apodization(**kwargs)
    # Divide by the number of points summed over, if average is True
    if average:
        values /= max(1, math.prod(spec.size(kwargs, dim) for dim in sum_dimensions))
    return values



from vbeam.fastmath import backend_manager

backend_manager.active_backend = "numpy"

import numpy as np
import pytest
from hypothesis import given
from vbeam_test_helpers.generators import sector_scans

from vbeam.scan import SectorScan, sector_scan


def _get_cartesian_bounds_brute_force(scan: SectorScan):
    """Just generate the points from the scan and calculate the minimum and maximum
    x-, y- (if 3D), and z-values."""
    points = scan.get_points()
    points -= scan.apex
    min_x, max_x = points[:, 0].min(), points[:, 0].max()
    min_z, max_z = points[:, 2].min(), points[:, 2].max()
    if scan.is_3d:
        min_y, max_y = points[:, 1].min(), points[:, 1].max()
        return (min_x, max_x, min_y, max_y, min_z, max_z)
    if scan.is_2d:
        return (min_x, max_x, min_z, max_z)


@given(sector_scans([200, 2]))
def test_cartesian_bounds_2D(scan: SectorScan):
    bounds_brute_force = _get_cartesian_bounds_brute_force(scan)
    bounds = scan.cartesian_bounds
    np.testing.assert_allclose(bounds_brute_force, bounds, rtol=1e-4, atol=1e-12)


@given(sector_scans([10, 10, 10]))
def test_cartesian_bounds_3D(scan: SectorScan):
    bounds_brute_force = _get_cartesian_bounds_brute_force(scan)
    bounds = scan.cartesian_bounds
    np.testing.assert_allclose(bounds_brute_force, bounds, rtol=1e-4, atol=1e-12)


regression_scans = [
    sector_scan(
        np.linspace(1.0, 0.0, 200),
        np.linspace(0.0, 1.0, 2),
        apex=np.array([0.1, 0.0, 0.2]),
    ),
    sector_scan(
        np.linspace(0.0, 2.0, 200),
        np.linspace(1.0, 0.0, 2),
        apex=np.array([0.1, 0.0, 0.2]),
    ),
]


@pytest.fixture(params=regression_scans)
def regression_scan(request):
    return request.param


def test_regression_scans(regression_scan):
    bounds_brute_force = _get_cartesian_bounds_brute_force(regression_scan)
    bounds = regression_scan.cartesian_bounds
    np.testing.assert_allclose(bounds_brute_force, bounds, rtol=1e-4, atol=1e-12)



"""Numpy-like API for multiple backends.

This module contains an abstraction over the operations needed by the vbeam project, 
that can be implemented by multiple backends. Current included implementations are 
Tensorflow, JAX, and Numpy. You may register custom backends using the register_backend
function.

fastmath uses a "BYOD" principle: "Bring Your Own Dependency". If you only want
to use Tensorflow, then you only need to install Tensorflow in your project (and not 
JAX, for example, or any other potential backend). Setting the active_backend to your
preferred backend should (hopefully) let you use vbeam as if it is written in that
backend.

fastmath is considered a separate, but internal project, meaning that you should not 
depend on fastmath in your own projects (besides setting the active backend). This is to 
ease the maintenance of the fastmath module, and only implement the operations needed 
internally by vbeam.

fastmath is inspired by Google's deep learning library TRAX:
https://github.com/google/trax/tree/master/trax/fastmath


The API can be imported via:
>>> from vbeam.fastmath import backend_manager

You may then set your preferred backend (default is "numpy") on the backend_manager:
>>> backend_manager.active_backend = "jax"  # ...or something else

You may also set the backend using a context manager:
>>> with backend_manager.using_backend("tensorflow"):
...     print(backend_manager.active_backend)

Note that setting the backend sets it globally and is not thread-safe.
"""

import contextlib
from typing import Callable, Dict, Optional

from vbeam.fastmath.backend import Backend
from vbeam.fastmath.included_backends import (
    get_best_available_backend,
    included_backends,
)

_registered_backends: Dict[str, Callable[[], Backend]] = included_backends


def register_backend(name: str, get_backend: Callable[[], Backend]):
    """Register a new Backend.

    Args:
      name: The unique name of the backend, e.g. "my_new_backend".
      get_backend: A function that takes no arguments and returns the Backend. It
        typically also dynamically loads the needed modules.

    Example:
    # In file my_project/my_backend.py
    class MyBackend(Backend):
        ...  # Implementations of all methods

    # In file my_project/__init__.py
    def get_my_backend():
        from my_project.my_backend import MyBackend
        return MyBackend()

    register_backend("my_backend", get_my_backend)

    # You may now use your custom backend:
    backend_manager.active_backend = "my_backend"
    """
    if name in _registered_backends:
        raise ValueError(
            f'The name of the backend must be unique: A backend with the name "{name}" has already been registered.'
        )
    _registered_backends[name] = get_backend


class BackendManager:
    _active_backend: Optional[str]
    _backends_cache: Dict[str, Backend] = {}

    def __init__(self, initial_active_backend: Optional[str] = None):
        self._active_backend = initial_active_backend

    @property
    def active_backend(self) -> Backend:
        if self._active_backend is None:
            self._active_backend = get_best_available_backend()

        backend = self._backends_cache.get(self._active_backend, None)
        if backend is None:
            get_backend = _registered_backends[self._active_backend]
            backend = get_backend()
            self._backends_cache[self._active_backend] = backend

        return backend

    @active_backend.setter
    def active_backend(self, new_backend: str):
        self._active_backend = new_backend
        self.active_backend  # Eagerly instantiate backend

    @contextlib.contextmanager
    def using_backend(self, new_backend: str):
        prev_backend = self._active_backend
        self._active_backend = new_backend
        try:
            yield
        finally:
            self._active_backend = prev_backend


def proxy_backend(backend_manager: BackendManager) -> Backend:
    """Return a proxy-backend that forwards all calls to the active backend."""

    class ProxyBackend(Backend):
        def __getattribute__(self, attr):
            return getattr(backend_manager.active_backend, attr)

    return ProxyBackend()


backend_manager = BackendManager()
numpy = proxy_backend(backend_manager)



from typing import Optional

import numpy

from vbeam.core import SpeedOfSound
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.interpolation import FastInterpLinspace
from vbeam.scan import Scan
from vbeam.util import ensure_2d_point
from vbeam.util.geometry.v2 import distance


@traceable_dataclass(
    data_fields=("values", "x_axis", "z_axis", "default_speed_of_sound"),
    aux_fields=("n_samples",),
)
class HeterogeneousSpeedOfSound(SpeedOfSound):
    """Sample the speed of sound between sender, point, and receiver position, and
    return the average.

    You should probably only ever use this if you are doing synthetic transmit aperture
    imaging (STAI). For more complex imaging setups, like focused, diverging, or plane
    wave imaging, this class is not appropriate. How should speed of sound be sampled
    for these setups? That's a difficult question :)"""

    values: np.ndarray
    x_axis: FastInterpLinspace
    z_axis: FastInterpLinspace
    n_samples: int
    default_speed_of_sound: float = 1540.0

    def average(
        self,
        sender_position: np.ndarray,
        point_position: np.ndarray,
        receiver_position: np.ndarray,
    ) -> float:
        sender_position = ensure_2d_point(sender_position)
        point_position = ensure_2d_point(point_position)
        receiver_position = ensure_2d_point(receiver_position)

        # Calculate distances to get correct weighting of the averages
        distance1 = distance(sender_position, point_position)
        distance2 = distance(point_position, receiver_position)
        total_distance = distance1 + distance2

        # Get the average speed of sounds
        average1 = self.average_between_two_points(sender_position, point_position)
        average2 = self.average_between_two_points(point_position, receiver_position)

        # Return the weighted average of the total distance
        return (average1 * distance1 + average2 * distance2) / total_distance

    def average_between_two_points(self, p1: np.ndarray, p2: np.ndarray) -> float:
        "Return the averaged sampled speed of sound between point ``p1`` and ``p2``."
        assert p1.shape == p2.shape == (2,), "Expected p1 and p2 to be 2D points."
        x1, z1 = p1[0], p1[1]
        x2, z2 = p2[0], p2[1]
        dx, dz = (x2 - x1) / self.n_samples, (z2 - z1) / self.n_samples
        # We use np.reduce instead of interp2d directly to allocate less GPU memory
        integrated_speed_of_sound = np.reduce(
            lambda carry, i: carry
            + FastInterpLinspace.interp2d(
                x1 + i * dx,
                z1 + i * dz,
                self.x_axis,
                self.z_axis,
                self.values,
                padding=self.default_speed_of_sound,
            ),
            np.arange(self.n_samples),
            np.array(0.0, dtype=self.values.dtype),
        )
        return integrated_speed_of_sound / self.n_samples

    @staticmethod
    def from_scan(
        scan: Scan,
        values: np.ndarray,
        n_samples: Optional[int] = None,
        default_speed_of_sound: float = 1540.0,
    ) -> "HeterogeneousSpeedOfSound":
        if not scan.is_2d:
            raise ValueError(
                "Only 2D scans are supported for HeterogeneousSpeedOfSound"
            )

        from_x, to_x, from_z, to_z = scan.cartesian_bounds
        n_x, n_z = scan.shape
        if n_samples is None:
            n_samples = int(numpy.ceil(numpy.sqrt(n_x**2 + n_z**2)))
        return HeterogeneousSpeedOfSound(
            values,
            x_axis=FastInterpLinspace(from_x, (to_x - from_x) / n_x, n_x),
            z_axis=FastInterpLinspace(from_z, (to_z - from_z) / n_z, n_z),
            n_samples=n_samples,
            default_speed_of_sound=default_speed_of_sound,
        )



from typing import Optional, Tuple, Union

from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util import ensure_2d_point
from vbeam.util.geometry.v2 import Line, distance

from .window import Window


@traceable_dataclass(("array_bounds", "window"))
class PlaneWaveTransmitApodization(Apodization):
    array_bounds: Tuple[np.ndarray, np.ndarray]
    window: Optional[Window] = None

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        # Set up the geometry. There is one line originating from each side of the
        # array, going in the direction of the transmitted wave. If the point is
        # outside of those lines, then it is weighted by 0.
        array_left, array_right = self.array_bounds

        # We are only supporting 2D points (for now).
        array_left = ensure_2d_point(array_left)
        array_right = ensure_2d_point(array_right)
        point_position = ensure_2d_point(point_position)

        # FIXME: There's something funky going on with the geometry code here 🤔 Angles
        # need to be inverted to make it work.
        azimuth = np.pi / 2 - wave_data.azimuth  # This is a hack to make it work.
        # Construct the lines going from either side of the array
        line_left = Line.with_angle(array_left, azimuth)
        line_right = Line.with_angle(array_right, azimuth)

        # The point is within the line if it has negative (to the left) signed distance
        # of one of the lines and positive (to the right) signed distance of the other.
        # Multiplying the signed distances gives a negative number if that's the case.
        # apodization_value is True if the point is within the lines, False otherwise.
        apodization_value = (
            line_left.signed_distance(point_position)
            * line_right.signed_distance(point_position)
        ) < 0

        if self.window is not None:
            # Apply the window function. It gradually goes to 0 as the point gets
            # closer to the edges of the apodization values.
            point_perp = Line.with_angle(point_position, azimuth + np.pi / 2)
            _, A = line_left.intersect(point_perp)
            _, B = line_right.intersect(point_perp)
            dist_A = distance(A, point_position)
            dist_B = distance(B, point_position)
            total_distance = dist_A + dist_B
            # p is 0 when the point is exactly between the two lines, and 0.5 when it is
            # on top of one of the lines.
            p = 0.5 - np.min(np.array([dist_A, dist_B])) / total_distance
            apodization_value *= self.window(p)

        # Ensure that it the returned value is a float.
        return apodization_value * 1.0


@traceable_dataclass(("window", "f_number"))
class PlaneWaveReceiveApodization(Apodization):
    window: Window
    f_number: Union[float, Tuple[float, float]]

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        receiver: ElementGeometry,
        wave_data: WaveData,
    ) -> float:
        f_number = (
            self.f_number
            if isinstance(self.f_number, tuple) and len(self.f_number) == 2
            else (self.f_number, self.f_number)
        )
        dist = point_position - receiver.position
        x_dist, y_dist, z_dist = dist[0], dist[1], dist[2]
        tan_theta = x_dist / z_dist
        tan_phi = y_dist / z_dist
        ratio_theta = np.abs(f_number[0] * tan_theta)
        ratio_phi = np.abs(f_number[1] * tan_phi)
        return np.array(
            self.window(ratio_theta) * self.window(ratio_phi),
            dtype="float32",
        )



from typing import TYPE_CHECKING, Optional, Tuple

from vbeam.fastmath import numpy as np
from vbeam.interpolation import FastInterpLinspace
from vbeam.util.arrays import grid

if TYPE_CHECKING:
    from vbeam.scan import SectorScan


def scan_convert(
    imaged_points: np.ndarray,
    sector_scan: "SectorScan",
    azimuth_axis: int = -2,
    depth_axis: int = -1,
    *,  # Remaining args must be passed by name (to avoid confusion)
    shape: Optional[Tuple[int, int]] = None,
    padding: Optional[np.ndarray] = 0.0,
):
    if shape is None:
        shape = sector_scan.shape

    # Get the points in the cartesian grid
    min_x, max_x, min_z, max_z = sector_scan.cartesian_bounds
    points = grid(
        np.linspace(min_x, max_x, shape[0]),
        np.linspace(min_z, max_z, shape[1]),
    )
    x, z = points[..., 0], points[..., 1]  # (Ignore y; scan_convert only supports 2D!)
    # and transform each point to polar coordinates.
    angles = np.arctan2(x, z)
    radii = np.sqrt(x**2 + z**2)

    # Interpolate the imaged points, sampled at the transformed points.
    return FastInterpLinspace.interp2d(
        angles,
        radii,
        FastInterpLinspace.from_array(sector_scan.azimuths),
        FastInterpLinspace.from_array(sector_scan.depths),
        imaged_points,
        azimuth_axis,
        depth_axis,
        padding=padding,
    )


def parse_axes(xyz):
    "Internal utility function for parsing axes. Handles both 2D and 3D scans."
    if len(xyz) == 3:
        x, y, z = xyz
    elif len(xyz) == 2:
        x, z = xyz
        y = None
    else:
        raise ValueError(
            f"Provide either x, y, and z (3D) or only x and z (2D). Got {len(xyz)} axes"
        )
    return x, y, z



from functools import reduce
from typing import Union

import tensorflow as tf

# Same as `import tensorflow.experimental.numpy as tnp`, but vscode doesn't recognize that for some reason.
import tensorflow._api.v2.experimental.numpy as tnp
from tensorflow.python.util import nest

from vbeam.fastmath.backend import Backend
from vbeam.fastmath.included_backends.tensorflow_backend.vmap import vmap
from vbeam.fastmath.traceable import is_traceable, is_traceable_dataclass

# Enable NumPy behavior on Tensors.
tnp.experimental_enable_numpy_behavior()
_traceable_classes_cache = {}


def swap_axes(t: tf.Tensor, ax1: int, ax2: int):
    return tf.transpose(
        t,
        [
            ax2 if dim_i == ax1 else ax1 if dim_i == ax2 else dim_i
            for dim_i in range(t.ndim)
        ],
    )


def make_type_work_with_tensorflow(T):
    if is_traceable(T):
        return Union[tf.Tensor, tf.experimental.BatchableExtensionType]

    if len(getattr(T, "__args__", ())) > 0:
        tf_types_args = [make_type_work_with_tensorflow(arg) for arg in T.__args__]
        return T.copy_with(tuple(tf_types_args))

    if is_traceable_dataclass(T):
        return tf.experimental.BatchableExtensionType

    # Fallback to tf.Tensor. This solves cases where the type is
    # float but value is scalar Tensor, which (I think) _should_
    # work, but doesn't. Tensorflow is quite strict about types here.
    if T is float:
        return Union[float, tf.Tensor]

    return T


class TensorflowBackend(Backend):
    @property
    def ndarray(self):
        return tnp.ndarray

    def zeros(self, shape, dtype=float):
        return tnp.zeros(shape, dtype)

    def ones(self, shape, dtype=float):
        return tnp.ones(shape, dtype)

    @property
    def pi(self):
        return tnp.pi

    def abs(self, x):
        return tnp.abs(x)

    def exp(self, x):
        return tnp.exp(x)

    def log(self, x):
        return tnp.log(x)

    def log10(self, x):
        return tnp.log10(x)

    def sin(self, x):
        return tnp.sin(x)

    def cos(self, x):
        return tnp.cos(x)

    def tan(self, x):
        return tnp.tan(x)

    def arcsin(self, x):
        return tnp.arcsin(x)

    def arccos(self, x):
        return tnp.arccos(x)

    def arctan2(self, y, x):
        return tnp.arctan2(y, x)

    def sqrt(self, x):
        return tnp.sqrt(x)

    def sign(self, x):
        return tnp.sign(x)

    def nan_to_num(self, x, nan=0.0):
        return tf.where(tnp.isnan(x), tf.ones_like(x) * nan, x)

    def min(self, a, axis=None):
        return tnp.min(a, axis=axis)

    def max(self, a, axis=None):
        return tnp.max(a, axis=axis)

    def argmin(self, a, axis=None):
        return tnp.argmin(a, axis=axis)

    def argmax(self, a, axis=None):
        return tnp.argmax(a, axis=axis)

    def minimum(self, a, b):
        return tnp.minimum(a, b)

    def maximum(self, a, b):
        return tnp.maximum(a, b)

    def sum(self, a, axis=None):
        return tnp.sum(a, axis=axis)

    def prod(self, a, axis=None):
        return tnp.prod(a, axis=axis)

    def mean(self, a, axis=None):
        return tnp.mean(a, axis=axis)

    def var(self, a, axis=None):
        return tnp.var(a, axis=axis)

    def cumsum(self, a, axis=None):
        return tnp.cumsum(a, axis=axis)

    def cross(self, a, b, axisa=-1, axisb=-1, axisc=-1, axis=None):
        return tnp.cross(a, b, axisa=axisa, axisb=axisb, axisc=axisc, axis=axis)

    def nansum(self, a, axis=None):
        return tnp.nansum(a, axis=axis)

    def histogram(self, a, bins=10, weights=None):
        return tnp.histogram(a, bins=bins, weights=weights)

    def array(self, x, dtype=None):
        return tnp.array(x, dtype=dtype)

    def transpose(self, a, axes=None):
        return tnp.transpose(a, axes=axes)

    def swapaxes(self, a, axis1, axis2):
        return tnp.swapaxes(a, axis1, axis2)

    def moveaxis(self, a, source, destination):
        return tnp.moveaxis(a, source, destination)

    def stack(self, arrays, axis=0):
        return tnp.stack(arrays, axis=axis)

    def tile(self, A, reps):
        return tnp.tile(A, reps)

    def concatenate(self, arrays, axis=0):
        # tnp.concatenate has slightly different behavior from numpy and JAX in that it
        # requires the arrays to be in a list. We can't create a list from a Tensor
        # while JIT-compiling, so we instead roll our own version.
        original_axis = axis
        if not isinstance(arrays, tf.Tensor):
            arrays = tnp.array(arrays)

        if axis < 0:
            axis = arrays.ndim - 1 + axis  # axis is negative so this is subtraction

        if arrays.ndim <= 1:
            raise ValueError("zero-dimensional arrays cannot be concatenated")
        if axis >= (arrays.ndim - 1) or axis < 0:
            raise ValueError(
                f"axis {original_axis} is out of bounds for array of dimension {arrays.ndim-1}"
            )

        arrays = swap_axes(arrays, 1, axis + 1)
        shape = arrays.shape
        arrays = tf.reshape(arrays, (shape[0] * shape[1], *shape[2:]))
        return swap_axes(arrays, 0, axis)

    def meshgrid(self, *xi, indexing="xy"):
        return tnp.meshgrid(*xi, indexing=indexing)

    def linspace(self, start, stop, num=50):
        return tnp.linspace(start, stop, num=num)

    def arange(self, start, stop=None, step=None):
        if step is None:
            step = 1
        return tnp.arange(start, stop, step)

    def expand_dims(self, a, axis):
        # Support multiple axes
        if not isinstance(axis, tuple):
            axis = (axis,)
        return reduce(tnp.expand_dims, axis, a)

    def ceil(self, x):
        return tnp.ceil(x)

    def floor(self, x):
        return tnp.floor(x)

    def modf(self, x):
        return tnp.modf(x)

    def round(self, x):
        return tnp.round(x)

    def clip(self, a, a_min, a_max):
        return tnp.clip(a, a_min, a_max)

    def where(self, condition, x, y):
        return tnp.where(condition, x, y)

    def select(self, condlist, choicelist, default=0):
        return tnp.select(condlist, choicelist, default)

    def logical_or(self, x1, x2):
        return tnp.logical_or(x1, x2)

    def logical_and(self, x1, x2):
        return tnp.logical_and(x1, x2)

    def squeeze(self, a, axis=None):
        return tnp.squeeze(a, axis=axis)

    def ravel(self, a):
        return tnp.ravel(a)

    def take(self, a, indices, axis=None):
        if is_traceable_dataclass(a.__class__):
            flattened_a = [
                tnp.take(a_tensor, indices, axis=axis)
                for a_tensor in nest.flatten(a, expand_composites=True)
            ]
            return nest.pack_sequence_as(a, flattened_a, expand_composites=True)
        else:
            return tnp.take(a, indices, axis=axis)

    def interp(self, x, xp, fp, left=None, right=None, period=None):
        # TODO: Implement tensorflow version
        return tnp.interp(x, xp, fp, left, right, period)

    def gather(self, a, indices):
        return tf.gather(a, indices)

    class add:
        @staticmethod
        def at(a, indices, b):
            if indices.dtype == bool:
                indices = tf.where(indices)
            else:
                # Flatten indices and updates
                indices, b = tnp.ravel(indices), tnp.ravel(b)
                # Explicitly convert negative indices to their corresponding positive ones
                neg1_indices = tf.where(indices < 0)
                adds = tf.repeat(a.shape[0], neg1_indices.shape[0])
                indices = tf.tensor_scatter_nd_add(
                    indices, neg1_indices, tf.cast(adds, indices.dtype)
                )
                # Add extra dimension so that tensorflow is happy :)
                indices = tnp.expand_dims(indices, -1)
            return tf.tensor_scatter_nd_add(a, indices, tf.cast(b, a.dtype))

    def jit(self, fun, static_argnums=None, static_argnames=None):
        if static_argnums is not None or static_argnames is not None:
            raise NotImplementedError(
                "Tensorflow backend currently doesn't support static_argnums or \
static_argnames"
            )
        return tf.function(fun, jit_compile=True)

    def vmap(self, fun, in_axes, out_axes=0):
        return tf.function(vmap(fun, in_axes, out_axes))

    def scan(self, f, init, xs):
        # TODO: perhaps use tf.scan instead? However, it has slightly different
        # semantics than jax.lax.scan which is what our backend is based on.
        carry = init
        ys = []
        for x in xs:
            carry, y = f(carry, x)
            ys.append(y)
        return carry, tnp.stack(ys)

    def as_traceable_dataclass_obj(self, obj, data_fields, aux_fields):
        # Make the object an instance of Tensorflow's ExtensionType so that it can be
        # traced by Tensorflow.
        #
        # Tensorflow doesn't need to differentiate between child attributes and
        # auxilliary data, so we can ignore the last two arguments: data_fields and
        # aux_fields.

        original_class = obj.__class__
        if original_class not in _traceable_classes_cache:
            # Make all the class attributes traceable.
            traceable_annotations = {
                k: make_type_work_with_tensorflow(v)
                for (k, v) in getattr(original_class, "__annotations__", {}).items()
            }
            # Create the new traceable version of the class
            TensorflowTraceable = type(
                original_class.__name__,
                (original_class, tf.experimental.BatchableExtensionType),
                {
                    "__qualname__": original_class.__qualname__,
                    "__annotations__": traceable_annotations,
                },
            )
            _traceable_classes_cache[original_class] = TensorflowTraceable

        # Overwrite the original class
        object.__setattr__(obj, "__class__", _traceable_classes_cache[original_class])
        return obj



"Interface for interpolating the :term:`signal` given a delay."

from abc import ABC, abstractmethod

from vbeam.fastmath import numpy as np


class InterpolationSpace1D(ABC):
    """An interface for interpolating data in 1D."""

    @abstractmethod
    def __call__(self, x: np.ndarray, fp: np.ndarray) -> np.ndarray:
        """Evaluate the points x on the discrete array fp.

        Any point in x that is outside of the range of fp is evaluated as zero."""
        ...

    @property
    @abstractmethod
    def start(self) -> float:
        ...

    @property
    @abstractmethod
    def end(self) -> float:
        ...



from vbeam.fastmath import Backend
from vbeam.util.vmap import vmap_all_except


def test_summing_one_axis_only(np: Backend):
    f = vmap_all_except(np.sum, 1)

    x = np.ones((2, 3, 4))
    result = f(x)
    assert result.shape == (2, 4), "Axis 1 is summed over and is removed from shape"
    assert (result == 3).all()


def test_f_produces_nd_array(np: Backend):
    @vmap_all_except(1)
    def f(x):
        return np.ones((3, 5, 6)) * np.sum(x)

    x = np.ones((2, 3, 4))
    result = f(x)
    assert result.shape == (2, 3, 5, 6, 4)
    assert (result == 3).all()


def test_only_one_axis_item_is_processed(np: Backend):
    x = np.zeros((2, 3, 4))

    @vmap_all_except(1)
    def f(x):
        assert x.shape == (3,)  # Only axis 1 is processed, rest is vectorized over
        return x

    result = f(x)
    assert result.shape == (2, 3, 4), "Shape is unchanged in no-op"
    assert (result == 0).all()



from typing import Callable, Literal, Optional, Tuple, Union, overload

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.scan.base import CoordinateSystem, Scan
from vbeam.scan.util import parse_axes
from vbeam.util.arrays import grid


@traceable_dataclass(("x", "y", "z"))
class LinearScan(Scan):
    x: np.ndarray
    y: Optional[np.ndarray]
    z: np.ndarray

    def get_points(self, flatten: bool = True) -> np.ndarray:
        shape = (self.num_points, 3) if flatten else (*self.shape, 3)
        y = np.array([0.0]) if self.y is None else self.y  # If the scan is 2D
        points = grid(self.x, y, self.z, shape=shape)
        return points

    def replace(
        self,
        x: Union[np.ndarray, Literal["unchanged"]] = "unchanged",
        y: Union[np.ndarray, Literal["unchanged"]] = "unchanged",
        z: Union[np.ndarray, Literal["unchanged"]] = "unchanged",
    ) -> "LinearScan":
        return LinearScan(
            x=x if x != "unchanged" else self.x,
            y=y if y != "unchanged" else self.y,
            z=z if z != "unchanged" else self.z,
        )

    def update(
        self,
        x: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        y: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        z: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    ) -> "LinearScan":
        return self.replace(
            x(self.x) if x is not None else "unchanged",
            y(self.y) if y is not None else "unchanged",
            z(self.z) if z is not None else "unchanged",
        )

    def resize(
        self,
        x: Optional[int] = None,
        y: Optional[int] = None,
        z: Optional[int] = None,
    ) -> "LinearScan":
        if y is not None and self.y is None:
            raise ValueError("Cannot resize y because it is not defined on this scan")
        return self.replace(
            np.linspace(self.x[0], self.x[-1], x) if x is not None else "unchanged",
            np.linspace(self.y[0], self.y[-1], y) if y is not None else "unchanged",
            np.linspace(self.z[0], self.z[-1], z) if z is not None else "unchanged",
        )

    @property
    def axes(self) -> Tuple[np.ndarray, ...]:
        if self.y is not None:
            return self.x, self.y, self.z
        else:
            return self.x, self.z

    @property
    def cartesian_bounds(self):
        return self.bounds  # LinearScan is already in cartesian coordinates :)

    @property
    def coordinate_system(self) -> CoordinateSystem:
        return CoordinateSystem.CARTESIAN

    def __repr__(self):
        return f"LinearScan(<shape={self.shape}>)"


@overload
def linear_scan(x: np.ndarray, z: np.ndarray) -> LinearScan:
    ...  # 2D scan


@overload
def linear_scan(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> LinearScan:
    ...  # 3D scan


def linear_scan(*xyz: np.ndarray) -> LinearScan:
    "Construct a linear scan. See LinearScan documentation for more details."
    x, y, z = parse_axes(xyz)
    return LinearScan(x, y, z)



from vbeam.core import ElementGeometry, TransmittedWavefront, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.geometry.v2 import distance


@traceable_dataclass(("base_wavefront", "base_sender"))
class REFoCUSWavefront(TransmittedWavefront):
    """A time-domain REFoCUS wavefront model.

    This class models the diverging spherical wave from a single transmitting /element/ 
    that was fired as part of an arbitrary focused transmit. It needs to know about the 
    original transmitted wave and the original sender (the point that the wave passed 
    through at time 0) in order to compensate for arbitrary transmit sequences. See 
    :meth:`__call__` for more details.
    """

    base_wavefront: TransmittedWavefront
    base_sender: ElementGeometry

    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        """Return the distance that the wave travels, starting from time 0, to when it
        reaches the point position.

        At its core, this is the distance between the sending element and the point
        position, but we also need to compensate for when the element fired in the
        transmit sequence. This compensation is the distance that the full modeled
        wavefront passed through the sending element."""
        focusing_compensation = self.base_wavefront(
            self.base_sender, sender.position, wave_data
        )
        return distance(sender.position, point_position) + focusing_compensation



from typing import Callable

from numpy import allclose, random
from vbeam.fastmath import Backend
from vbeam.util.coordinate_systems import as_cartesian, as_polar


def test_inverse_conversions(np: Backend, jit_able: Callable[[Callable], Callable]):
    pol2cart2pol = jit_able(lambda points: as_polar(as_cartesian(points)))
    cart2pol2cart = jit_able(lambda points: as_cartesian(as_polar(points)))
    points = random.random((10, 3))
    assert allclose(pol2cart2pol(points), points)
    assert allclose(cart2pol2cart(points), points)



"""Wavefront models that return the distance that a wave has traveled. 
:class:`TransmittedWavefront` models the transmitted wave while 
:class:`ReflectedWavefront` models a reflected/backscattered wave.

Wavefront models return a distance in meters. This way they are decoupled from the 
speed of sound of the medium."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Union

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.util.geometry.v2 import distance

from .element_geometry import ElementGeometry
from .wave_data import WaveData


class TransmittedWavefront(ABC):
    """The base class for defining wavefront models of transmitted waves.

    :class:`TransmittedWavefront` models are used to calculate the *distance (in
    meters)* from a sending element to a point. This distance, along with the distance
    returned from :class:`ReflectedWavefront` and when divided by a speed-of-sound, can
    be used to delay the signal of a receiving element.

    Most :class:`TransmittedWavefront` models return a single distance for a wave.
    However, it may optionally return more than one distance value in the form of a
    :class:`MultipleTransmitDistances` object. This is useful for wavefront models that
    sample multiple distances for a single transmit.

    See also:
        :class:`ReflectedWavefront`
        :class:`MultipleTransmitDistances`
        :func:`~vbeam.core.kernels.signal_for_point`

    Examples:
        We can create a simple synthetic transmit aperture (STA) wavefront model where
        a single sender element sends out a spherical wavefront. The returned distance
        is then simply the euclidean distance between the sender and point. To
        illustrate the use of :class:`MultipleTransmitDistances`, we also add a small
        smoothing to the distance by averaging with two small delays (+/- 0.1
        millimeters):

        >>> class TransmittedWavefront(Wavefront):
        ...     def __call__(self, sender, point_position, wave_data):
        ...         # Simply get the distance from sender to point (assume STA)
        ...         dist = distance(sender.position, point_position)
        ...         # Smooth the dist by averaging with the two neighboring distances
        ...         dist = jnp.array([dist - 1e-4, dist, dist + 1e-4])
        ...         # Weights for the three different delayed signal samples
        ...         weights = jnp.array([0.2, 0.6, 0.2])
        ...         return MultipleTransmitDistances(
        ...             dist,
        ...             # The samples values are aggregated by a weighted sum
        ...             lambda samples: np.sum(samples * weights),
        ...         )
    """

    @abstractmethod
    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> Union[float, "MultipleTransmitDistances"]:
        """Return the *distance (in meters)* from the sender element to the point for a transmit.

        May optionally return a :class:`MultipleTransmitDistances` object for cases
        where the wavefront model samples multiple distances for a single transmit.

        See also:
            :class:`MultipleTransmitDistances`
            :func:`~vbeam.core.kernels.signal_for_point`
        """


@traceable_dataclass()
class ReflectedWavefront:
    """The base class for defining wavefront models of reflected/backscattered waves.

    This is usually just the euclidian distance as there are limits to what we can
    model.

    See also:
        :class:`TransmittedWavefront`
        :func:`~vbeam.core.kernels.signal_for_point`"""

    def __call__(self, point_position: np.ndarray, receiver: ElementGeometry) -> float:
        return distance(point_position, receiver.position)


@dataclass
class MultipleTransmitDistances:
    """Multiple distance values returned from a :class:`TransmittedWavefront`.

    Some more advanced :class:`TransmittedWavefront` models may return multiple
    distances for a transmitted wave. In this case, each returned distance will be used
    to delay the element signals, and the delayed samples will be combined using the
    function set in :attr:`aggregate_samples`. The :attr:`aggregate_samples` function
    may for example weight the delayed signals differently before summing.

    See reference to :class:`MultipleTransmitDistances` in
    :func:`~vbeam.core.kernels.signal_for_point` for implementation details.

    Mathematical operators applied to the :class:`MultipleTransmitDistances` object
    will apply them to the :attr:`values` attribute as if it was just a numpy array.

    Attributes:
        values (np.ndarray): The distance values that will be used to delay the element signals.
        aggregate_samples (Callable[[np.ndarray, np.ndarray], np.ndarray]): A function
            that will be used to combine the delayed samples into one value, for
            example as a weighted sum. If None, the samples will be averaged.

    See also:
        :func:`~vbeam.core.kernels.signal_for_point`
        :class:`TransmittedWavefront`
    """

    values: np.ndarray
    aggregate_samples: Callable[[np.ndarray, np.ndarray], np.ndarray] = None

    def __post_init__(self):
        # If no aggregate function is set, just average the samples
        if self.aggregate_samples is None:
            self.aggregate_samples = np.mean

    # Mathematical operators applied to the :class:`MultipleTransmitDistances` object
    # will apply them to the ``values`` attribute as if it was just a numpy array.
    def __truediv__(self, other) -> np.ndarray:
        return self.values / other

    def __add__(self, other) -> np.ndarray:
        return self.values + other



from typing import Optional, Sequence, Tuple

from spekk import Spec

from vbeam.apodization.util import get_apodization_values
from vbeam.core import Apodization, ElementGeometry, WaveData
from vbeam.fastmath import backend_manager
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass
from vbeam.scan.advanced.base import ExtraDimsScanMixin, WrappedScan
from vbeam.scan.base import Scan
from vbeam.util.transformations import *
from vbeam.util.vmap import vmap_all_except


def get_point_indices_for_transmits(
    apodization_values: np.ndarray,
    threshold: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """Return the indices (and indices mask which is explained further
    down) for the points in an image that are to be imaged for a given transmit. Which
    points that are selected to be imaged are determined by the given apodization object
    and threshold.

    The shape of the returned arrays is (num_transmits, num_points_per_transmit).

    If the apodization value for a given point is greater than threshold then that point
    will be imaged for the given transmit.

    ## Explanation of indices_mask:
    It is important that the size of a dimension in an array is consistent in order to
    compile it using XLA. However, the number of points to be imaged may differ between
    transmits — some transmits may have fewer or more points in focus than other
    transmits.

    We add extra "padding"-points to transmits that have fewer points than the maximum
    number of points for a transmit. This padding has to be removed afterwards, so we
    also return indices_mask which is 0 for points that are added as padding.

    For example, if the number of points for 3 different transmits were 4, 6, and 5,
    then padding would be added so that 6 points were imaged for each transmit. The
    indices_mask would be [1,1,1,1,0,0], [1,1,1,1,1,1], , [1,1,1,1,1,0], respectively.
    Notice how the first 4 values in the mask for the first transmit are 1s while the
    two remaining are 0s.
    """

    # Apply thresholding to get the masks.
    masks = apodization_values >= threshold

    # Get the maximum number of points to image.
    num_focused_points = np.sum(masks, -1)
    max_num_focused_points = np.max(num_focused_points)

    # We use Numpy for performing the actual masking because that is hard to do on GPUs
    with backend_manager.using_backend("numpy"):

        def mask1(mask):
            masked_indices = np.arange(mask.size)[mask]
            num_masked_indices = np.sum(mask)
            # We set padding-indices to -1 (these will be ignored because of indices_mask)
            padding = np.ones((max_num_focused_points - num_masked_indices,)) * -1
            return np.concatenate([masked_indices, padding])

        m_dimensions = [f"dim{i}" for i in range(masks.ndim)]
        get_indices = compose(
            mask1,
            # Loop over all dimensions except the last one, which is expected to be the
            # points dimension.
            *[ForAll(dim) for dim in m_dimensions[:-1]],
        ).build(Spec({"mask": m_dimensions}))
        indices = get_indices(mask=masks)

    indices = indices.astype("int32")
    indices_mask = indices != -1
    return np.array(indices), np.array(indices_mask)


def recombine1(
    image: np.ndarray,
    imaged_points: np.ndarray,
    indices: np.ndarray,
    indices_mask: np.ndarray,
    points_axis: int,
):
    @vmap_all_except(points_axis)
    def recombine(imaged_points: np.ndarray) -> np.ndarray:
        return np.add.at(image, indices, imaged_points * indices_mask)

    return recombine(imaged_points)


@traceable_dataclass(
    ("_points", "_indices", "_indices_mask"),
    (
        "base_scan",
        "apodization",
        "sender",
        "receiver",
        "wave_data",
        "spec",
        "dimensions",
        "threshold",
    ),
)
class ApodizationFilteredScan(WrappedScan, ExtraDimsScanMixin):
    def __init__(
        self,
        base_scan: Scan,
        apodization: Apodization,
        sender: ElementGeometry,
        receiver: ElementGeometry,
        wave_data: WaveData,
        spec: Spec,
        dimensions: Sequence[str],
        threshold: float,
        _points: Optional[np.ndarray] = None,
        _indices: Optional[np.ndarray] = None,
        _indices_mask: Optional[np.ndarray] = None,
    ):
        # We need to store these values in case we need to recompute the indices.
        # For example, if a user resizes the scan then the indices must be recomputed.
        self._base_scan = base_scan
        self.apodization = apodization
        self.sender = sender
        self.receiver = receiver
        self.wave_data = wave_data
        self.spec = spec
        self.dimensions = dimensions
        self.threshold = threshold

        # If any of these values are None it means we are constructing a new instance
        # (contra just copying an existing one) and we need to compute the indices.
        self._points = _points
        self._indices = _indices
        self._indices_mask = _indices_mask
        if _points is None or _indices is None or _indices_mask is None:
            self._recompute_indices()

    def get_points(self, flatten: bool = True) -> np.ndarray:
        if not flatten:
            raise ValueError(
                f"{self.__class__.__name__} only supports getting flattened points."
            )
        return self._points

    def unflatten(
        self,
        imaged_points: np.ndarray,
        transmits_axis: int,
        points_axis: int,
    ) -> np.ndarray:
        recombine = np.vmap(
            recombine1,
            [None, transmits_axis, 0, 0, None],
        )
        image = np.zeros((self.base_scan.num_points,), dtype=imaged_points.dtype)
        recombined_points = recombine(
            image,
            imaged_points,
            self._indices,
            self._indices_mask,
            points_axis - 1 if points_axis > transmits_axis else points_axis,
        )
        recombined_points = np.sum(recombined_points, 0)
        if points_axis > transmits_axis:
            points_axis -= 1
        return self.base_scan.unflatten(recombined_points, points_axis)

    @property
    def num_points(self) -> Tuple[int, ...]:
        return self._indices.shape[-1]

    @WrappedScan.base_scan.setter
    def base_scan(self, new_base_scan):
        WrappedScan.base_scan.fset(self, new_base_scan)
        self._recompute_indices()

    def _recompute_indices(self):
        points = self.base_scan.get_points()
        apodization_values = get_apodization_values(
            self.apodization,
            self.sender,
            points,
            self.receiver,
            self.wave_data,
            self.spec,
            self.dimensions,
            average_overlap=True,
        )

        points = self.base_scan.get_points()
        self._indices, self._indices_mask = get_point_indices_for_transmits(
            apodization_values,
            self.threshold,
        )
        self._points = points[self._indices]

    def copy(self):
        return ApodizationFilteredScan(
            self.base_scan,
            self.apodization,
            self.sender,
            self.receiver,
            self.wave_data,
            self.threshold,
            self._points,
            self._indices,
            self._indices_mask,
        )

    @staticmethod
    def from_setup(
        setup,
        dimensions: Sequence[str],
        threshold: float = 0.1,
        scan: Optional[Scan] = None,
    ) -> "ApodizationFilteredScan":
        from vbeam.data_importers import SignalForPointSetup

        if not isinstance(setup, SignalForPointSetup):
            raise TypeError(
                f"Expected a SignalForPointSetup, but type(setup) is {type(setup)!r}."
            )
        return ApodizationFilteredScan(
            scan if scan is not None else setup.scan,
            setup.apodization,
            setup.sender,
            setup.receiver,
            setup.wave_data,
            setup.spec,
            dimensions,
            threshold,
        )

    @property
    def flattened_points_dimensions(self) -> Sequence[str]:
        return self.dimensions

    @property
    def required_dimensions_for_unflatten(self) -> Sequence[str]:
        return self.dimensions



"""Differential 2D curves and relevant functions.

It's all in this one module because it's all so closely related."""


from abc import ABC, abstractmethod
from typing import Optional, Tuple, Union

from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass

####
# Functions for calculating intersections.
# Note, not all combinations of curves are supported.


def intersect_line_line(
    line1: "Line", line2: "Line"
) -> Tuple[int, Tuple[np.ndarray, np.ndarray]]:
    num_intersections = np.where(np.cross(line1.direction, line2.direction) == 0, 0, 1)
    return num_intersections, (
        line1.anchor
        + np.cross(line2.anchor - line1.anchor, line2.direction)
        / np.cross(line1.direction, line2.direction)[..., None]
        * line1.direction
    )


def intersect_circle_line(
    circle: "Circle", line: "Line"
) -> Tuple[int, Tuple[np.ndarray, np.ndarray]]:
    # http://paulbourke.net/geometry/circlesphere/
    a = np.sum(line.direction**2)
    b = 2 * np.sum((line.direction * (line.anchor - circle.center)))
    c = (
        np.sum(circle.center**2)
        + np.sum(line.anchor**2)
        - 2 * np.sum(circle.center * line.anchor)
        - circle.radius**2
    )

    discriminant = b**2 - 4 * a * c + 0j
    u1 = (-b + np.sqrt(discriminant)) / (2 * a)
    u2 = (-b - np.sqrt(discriminant)) / (2 * a)

    num_intersections = 0
    num_intersections = np.where(discriminant == 0, 1, num_intersections)
    num_intersections = np.where(discriminant > 0, 2, num_intersections)
    return num_intersections, np.array([line(u1), line(u2)]).T


####
# Some helper functions


def distance(point1: np.ndarray, point2: Optional[np.ndarray] = None):
    diff = point1 if point2 is None else point2 - point1
    return np.sqrt(np.sum(diff**2, axis=-1))


def rotate(
    point: np.ndarray, radians: float, anchor: Optional[np.ndarray] = None
) -> np.ndarray:
    """Rotate a point around an anchor point."""
    if anchor is None:
        anchor = np.array([0, 0])
    return (
        np.array(
            [
                np.cos(radians) * (point[0] - anchor[0])
                - np.sin(radians) * (point[1] - anchor[1]),
                np.sin(radians) * (point[0] - anchor[0])
                + np.cos(radians) * (point[1] - anchor[1]),
            ]
        )
        + anchor
    )


####
# Curve classes


class Curve(ABC):
    """A 2D curve. See method docstrings for details."""

    @abstractmethod
    def __call__(self, t: Union[float, np.ndarray]) -> Tuple[float, float]:
        """Evaluate the curve at a given parameter value.

        For example, evaluating a circle at t=0 gives the point at angle 0, t=pi/2 gives
        the point at 90 degrees angle, etc.

        Differentiating this, e.g. via jax.jvp, gives the tangent of the curve at a
        given point."""

    @abstractmethod
    def signed_distance(self, point: np.ndarray) -> np.ndarray:
        """Return the signed distance from a point to the nearest point on the curve.

        The sign of the returned value gives information about which side of the curve
        the point is on. The sign is positive on the right-hand side and negative on the
        left-hand side of the curve. For a circle, the distance is positive on the
        outside, and negative on the inside."""

    @abstractmethod
    def intersect(self, other: "Curve") -> Tuple[int, Tuple[np.ndarray, np.ndarray]]:
        """Return the number of intersections between this curve and another curve and
        the (potentially none) intersection points."""


@traceable_dataclass(("anchor", "direction"))
class Line(Curve):
    anchor: np.ndarray
    direction: np.ndarray

    def __post_init__(self):
        # Normalize direction vector (tangent)
        self.direction = self.direction / distance(self.direction)

    def __call__(self, t: Union[float, np.ndarray]) -> Tuple[float, float]:
        if np.is_ndarray(t) and t.ndim != 0:
            points = np.moveaxis(self.anchor + t[..., None] * self.direction, -1, 0)
        else:
            points = self.anchor + t * self.direction
        return points

    def signed_distance(self, point: np.ndarray) -> float:
        return np.cross(point - self.anchor, self.direction)

    @staticmethod
    def passing_through(point1: np.ndarray, point2: np.ndarray) -> "Line":
        return Line(point1, point2 - point1)

    @staticmethod
    def with_angle(anchor: np.ndarray, angle: float) -> "Line":
        return Line(anchor, np.array([np.cos(angle), np.sin(angle)]))

    def intersect(self, other: Curve) -> Tuple[int, Tuple[np.ndarray, np.ndarray]]:
        if isinstance(other, Line):
            return intersect_line_line(self, other)
        else:
            raise NotImplementedError

    @property
    def angle(self) -> float:
        return np.arctan2(self.direction[1], self.direction[0])

    @property
    def normal(self) -> np.ndarray:
        return np.array([-self.direction[1], self.direction[0]])


@traceable_dataclass(("center", "radius"))
class Circle(Curve):
    center: np.ndarray
    radius: float

    def __call__(self, t: float) -> Tuple[float, float]:
        center = self.center
        if np.is_ndarray(t) and t.ndim != 0:
            center = self.center[:, None]
        return np.array([np.cos(t), np.sin(t)]) * self.radius + center

    def signed_distance(self, point: np.ndarray) -> np.ndarray:
        return distance(point, self.center) - self.radius

    def intersect(self, other: Curve) -> Tuple[int, Tuple[np.ndarray, np.ndarray]]:
        if isinstance(other, Line):
            return intersect_circle_line(self, other)
        else:
            raise NotImplementedError


@traceable_dataclass(("f1", "f2", "d"))
class Ellipse(Curve):
    """An ellipse defined by two focus points and the summed distance from each to
    points on the ellipse.

    In other words: if p is a point on the ellipse, then d = distance(p, f1) + distance(p, f2)

    An ellipse is just a transformed unit circle, and that's how we implement Curve
    behavior for ellipses. As a consequence, the signed distance function is not yet
    implemented because transforming space also transforms the distance function. We
    should also not depend on the derivative of the __call__ method to compute a
    normalized tangent."""

    f1: np.ndarray  # Focus point 1
    f2: np.ndarray  # Focus point 2
    d: float  # If p is a point on the ellipse, then d = distance(p, f1) + distance(p, f2)

    def __call__(self, t: np.ndarray) -> np.ndarray:
        # An ellipse is just a transformed circle
        circle = Circle(np.array([0, 0]), 1)
        circle_transform = self._circle_transform
        if np.is_ndarray(t) and t.ndim != 0:
            circle_transform = np.vmap(circle_transform, [1])
        return circle_transform(circle(t)).T

    def signed_distance(self, point: np.ndarray) -> np.ndarray:
        # Not implemented yet. Need to think about how to normalize the distance since
        # the space is transformed.
        raise NotImplementedError

    def intersect(self, other: Curve) -> Tuple[int, Tuple[np.ndarray, np.ndarray]]:
        if isinstance(other, Line):
            other = Line.passing_through(
                self._circle_undo_transform(other(0)),
                self._circle_undo_transform(other(1)),
            )
            circle = Circle(np.array([0, 0]), 1)
            num_intersections, intersections = intersect_circle_line(circle, other)
            i1, i2 = intersections.real.T
            return (
                num_intersections,
                np.array([self._circle_transform(i1), self._circle_transform(i2)]).T,
            )
        else:
            raise NotImplementedError

    def _circle_rotation(self) -> float:
        return np.arctan2(self.f2[1] - self.f1[1], self.f2[0] - self.f1[0])

    def _circle_scale(self) -> np.ndarray:
        a = self.d / 2
        b = np.sqrt(a**2 - distance(self.f2, self.f1) ** 2 / 4)
        return np.array([a, b])

    def _circle_translation(self) -> np.ndarray:
        return (self.f1 + self.f2) / 2

    def _circle_transform(self, point: np.ndarray) -> np.ndarray:
        point *= self._circle_scale()
        point = rotate(point, self._circle_rotation())
        point += self._circle_translation()
        return point

    def _circle_undo_transform(self, point: np.ndarray) -> np.ndarray:
        point -= self._circle_translation()
        point = rotate(point, -self._circle_rotation())
        point /= self._circle_scale()
        return point



from vbeam.core import ElementGeometry, TransmittedWavefront, WaveData
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass()
class PlaneWavefront(TransmittedWavefront):
    def __call__(
        self,
        sender: ElementGeometry,
        point_position: np.ndarray,
        wave_data: WaveData,
    ) -> float:
        diff = point_position - sender.position
        x, y, z = diff[0], diff[1], diff[2]
        return (
            x * np.sin(wave_data.azimuth) * np.cos(wave_data.elevation)
            + y * np.sin(wave_data.elevation)
            + z * np.cos(wave_data.azimuth) * np.cos(wave_data.elevation)
        )



from typing import Callable

from numpy import allclose
from vbeam.fastmath import Backend
from vbeam.interpolation import FastInterpLinspace


def test_interp2d(np: Backend, jit_able: Callable[[Callable], Callable]):
    interp_x = FastInterpLinspace(10, 1, 2)  # x coords = [10, 11]
    interp_y = FastInterpLinspace(20, 1, 2)  # y coords = [20, 21]
    values = np.array([[10, 20], [30, 40]])

    @jit_able
    def i2d(eval_x, eval_y):
        return FastInterpLinspace.interp2d(eval_x, eval_y, interp_x, interp_y, values)

    # Interpolate x-axis
    assert allclose(i2d(np.array([9.5]), np.array([20.0])), [0.0]), "Outside of bounds"
    assert allclose(i2d(np.array([10.0]), np.array([20.0])), [10.0])
    assert allclose(i2d(np.array([10.5]), np.array([20.0])), [20.0])
    assert allclose(i2d(np.array([11.0]), np.array([20.0])), [30.0])
    assert allclose(i2d(np.array([11.5]), np.array([20.0])), [0.0]), "Outside of bounds"
    # Interpolate y-axis
    assert allclose(i2d(np.array([10.0]), np.array([19.5])), [0.0]), "Outside of bounds"
    assert allclose(i2d(np.array([10.0]), np.array([20.0])), [10.0])
    assert allclose(i2d(np.array([10.0]), np.array([20.5])), [15.0])
    assert allclose(i2d(np.array([10.0]), np.array([21.0])), [20.0])
    assert allclose(i2d(np.array([10.0]), np.array([21.5])), [0.0]), "Outside of bounds"
    # Interpolate exact middle
    assert allclose(i2d(np.array([10.5]), np.array([20.5])), [np.mean(values)])


def test_interp2d_nonscalars(np: Backend, jit_able: Callable[[Callable], Callable]):
    interp_x = FastInterpLinspace(10, 1, 2)  # x coords = [10, 11]
    interp_y = FastInterpLinspace(20, 1, 2)  # y coords = [20, 21]
    # Each value consists of 3 elements instead of just being scalars
    values = np.array(
        [
            [[1, 2, 3], [11, 12, 13]],
            [[21, 22, 23], [31, 32, 33]],
        ]
    )

    @jit_able
    def i2d(eval_x, eval_y):
        return FastInterpLinspace.interp2d(eval_x, eval_y, interp_x, interp_y, values)

    assert allclose(
        i2d(np.array([10]), np.array([20])),
        np.array([[1, 2, 3]]),
    ), "Smallest x and smallest y gets first element"
    assert allclose(
        i2d(np.array([11]), np.array([21])),
        np.array([[31, 32, 33]]),
    ), "Biggest x and Biggest y gets last element"
    assert allclose(
        i2d(np.array([10.5]), np.array([20])),
        np.array([[(1 + 21) / 2, (2 + 22) / 2, (3 + 23) / 2]]),
    ), "Interpolation is element-wise between values"
    assert allclose(
        i2d(np.array([10.0, 10.5, 11.0]), np.array([20.0, 20.5, 21.0])),
        np.array([[1, 2, 3], [16, 17, 18], [31, 32, 33]]),
    ), "We can interpolate multiple values at once"



from vbeam.scan.advanced.apodization_filtered_scan import ApodizationFilteredScan
from vbeam.scan.advanced.base import ExtraDimsScanMixin, WrappedScan
from vbeam.scan.advanced.scan_converted_sector_scan import ScanConvertedSectorScan

__all__ = [
    "ApodizationFilteredScan",
    "ExtraDimsScanMixin",
    "WrappedScan",
    "ScanConvertedSectorScan",
]



from dataclasses import dataclass
from functools import partial
from typing import Callable, Optional, Sequence, Union

from spekk import Spec

from vbeam.fastmath import numpy as np
from vbeam.util.transformations import *


def vmap_all_except(f: Union[Callable, int], axis: Optional[int] = None):
    # See if we're being called as @vmap_all_except or @vmap_all_except().
    if axis is None:
        # axis being None indicates we are being called as @vmap_all_except().
        # Then f must be the axis (i.e.: an int).
        if not isinstance(f, int):
            raise ValueError("axis must be specified when using @vmap_all_except().")
        return partial(vmap_all_except, axis=f)
    elif f is None:
        # axis may also have been specified as a keyword argument.
        if not isinstance(axis, int):
            raise ValueError("axis must be an int when using @vmap_all_except().")
        return partial(vmap_all_except, axis=f)

    def wrapped(x: np.ndarray):
        if x.ndim == 1:
            return f(x)

        nonlocal axis
        if axis < 0:
            axis += x.ndim
        dims = [f"dim{i}" for i in range(x.ndim)]
        vmapped_f = compose(
            lambda x: f(x),
            *[ForAll(dim) for dim in reversed(dims) if dim != f"dim{axis}"],
        )
        vmapped_f = vmapped_f.build(Spec({"x": dims}))
        result = vmapped_f(x=x)
        f_ndim = result.ndim - x.ndim + 1
        if f_ndim > 0:
            # f returned an array with at least 1 dimension. Transpose the result such
            # that those dimensions are at the given axis.
            result = np.transpose(
                result,
                [
                    *range(axis),
                    *range(result.ndim - f_ndim, result.ndim),
                    *range(axis, result.ndim - f_ndim),
                ],
            )

        return result

    return wrapped


def apply_binary_operation_across_axes(
    a: np.ndarray,
    b: np.ndarray,
    binary_operation: Callable[[np.ndarray, np.ndarray], np.ndarray],
    axes: Sequence[int],
) -> np.ndarray:
    """Apply a binary operation to two arrays, where the second array is potentially
    broadcasted and transposed to match the shape and axis-ordering of the first array.

    Args:
        a: The main array.
        b: The second array with potentially fewer axes than ``a``.
        binary_operation: A function that takes two arrays of the same shape and
            returns an array of the same shape.
        axes: The axes of `a` that corresponds to the axes of `b`.

    Let's say we have two arrays, ``a`` and ``b``. They share a dimension, in this
    case, the first dimension of ``b`` corresponds to the second dimension of ``a``:

    >>> dim0, dim1, dim2 = 2, 3, 4
    >>> a = np.ones([dim0, dim1, dim2])
    >>> b = np.ones([dim1])

    Now we want to add the arrays together, but ``a+b`` would raise an error since they
    are not broadcastable. With :func:`apply_binary_operation_across_axes` we can
    specify which axes of ``b`` should be used when applying the binary operation:

    >>> import operator
    >>> result = apply_binary_operation_across_axes(a, b, operator.add, [1])
    >>> result.shape
    (2, 3, 4)

    It also automatically transposes ``b`` to match the axis ordering of ``a``:

    >>> a = np.ones([dim0, dim1, dim2]) # <- dim2 comes AFTER dim1 in a!
    >>> b = np.ones([dim2, dim1])       # <- dim2 comes BEFORE dim1 in b!
    >>> result = apply_binary_operation_across_axes(a, b, operator.add, [2, 1])
    >>> result.shape
    (2, 3, 4)
    """
    if not isinstance(axes, Sequence):
        raise TypeError(
            f"axes must be a sequence of ints, but got {type(axes)}.",
        )
    if len(axes) != b.ndim:
        raise ValueError(
            f"The number of axes must match the number of dimensions of b. Expected {b.ndim} axes, but got only {len(axes)}.",
        )
    if any(axis < 0 or axis >= a.ndim for axis in axes):
        raise ValueError(
            f"All given axes must be between 0 and {a.ndim-1} since they refer to axes in a, but got {axes}.",
        )

    # Make sure that the axes of b are in the correct order so that we don't have to
    # transpose them later:
    ordering = sorted(enumerate(axes), key=lambda x: x[1])
    axes = [axis for _, axis in ordering]  # Re-order axis indices
    b = np.transpose(b, [i for i, _ in ordering])  # Re-order axes of b

    # Create a Spec for the dimensions of the data (names of dimensions doesn't matter):
    a_dims = [f"dim{axis}" for axis in range(a.ndim)]
    b_dims = [f"dim{axis}" for axis in axes]
    spec = Spec({"a": a_dims, "b": b_dims})

    # Transform f so that it works for the given data:
    tf = compose(
        Specced(
            lambda a, b: binary_operation(a, b),
            lambda input_spec: input_spec["a"],
        ),
        *[ForAll(dim) for dim in a_dims if dim not in b_dims],
        Apply(np.transpose, [Axis(dim, keep=True) for dim in a_dims]),
    ).build(spec)
    return tf(a=a, b=b)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from vbeam.core.wavefront import ReflectedWavefront
from vbeam.wavefront.focused import FocusedHybridWavefront, FocusedSphericalWavefront
from vbeam.wavefront.plane import PlaneWavefront
from vbeam.wavefront.refocus import REFoCUSWavefront
from vbeam.wavefront.stai import STAIWavefront
from vbeam.wavefront.unified import UnifiedWavefront

__all__ = [
    "ReflectedWavefront",
    "FocusedHybridWavefront",
    "FocusedSphericalWavefront",
    "PlaneWavefront",
    "REFoCUSWavefront",
    "STAIWavefront",
    "UnifiedWavefront",
]



from abc import ABC, abstractmethod
from typing import Callable, Literal, Optional, Sequence, Tuple, TypeVar, Union

from vbeam.fastmath import numpy as np
from vbeam.scan.base import CoordinateSystem, Scan

TSelf = TypeVar("TSelf", bound="WrappedScan")


def _wrap_with_new_base_scan(wrapped_scan: TSelf, new_base_scan: Scan) -> TSelf:
    new_copy = wrapped_scan.copy()
    new_copy.base_scan = new_base_scan
    return new_copy


class WrappedScan(Scan, ABC):
    """A scan that wraps another scan (the base_scan).

    This is useful for composing scans to make more advanced scans."""

    # The wrapped scan (accessed via the base_scan property).
    _base_scan: Scan

    def replace(
        self: TSelf,
        # "unchanged" means that the axis will not be changed.
        **kwargs: Union[np.ndarray, None, Literal["unchanged"]],
    ) -> TSelf:
        return _wrap_with_new_base_scan(self.base_scan.replace(**kwargs))

    def update(
        self: TSelf, **kwargs: Optional[Callable[[np.ndarray], np.ndarray]]
    ) -> TSelf:
        return _wrap_with_new_base_scan(self.base_scan.update(**kwargs))

    def resize(self: TSelf, kwargs: Optional[int]) -> TSelf:
        return _wrap_with_new_base_scan(self.base_scan.resize(**kwargs))

    @property
    def axes(self) -> Tuple[np.ndarray, ...]:
        return self.base_scan.axes

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.base_scan.shape

    @property
    def bounds(self) -> np.ndarray:
        return self.base_scan.bounds

    @property
    def cartesian_bounds(self) -> np.ndarray:
        return self.base_scan.cartesian_bounds

    @property
    def coordinate_system(self) -> CoordinateSystem:
        return self.base_scan.coordinate_system

    @property
    def is_2d(self) -> bool:
        return self.base_scan.is_2d

    @property
    def is_3d(self) -> bool:
        return self.base_scan.is_3d

    @property
    def base_scan(self):
        "Get the wrapped scan."
        return self._base_scan

    @base_scan.setter
    def base_scan(self, new_base_scan):
        "Set the wrapped scan."
        # You should override this setter in your subclass if this changes something.
        self._base_scan = new_base_scan

    @abstractmethod
    def copy(self: TSelf) -> TSelf:
        "Return a copy of the scan."

    def __getattr__(self, name):
        "Forward any other attribute access to the base_scan."
        try:
            return getattr(self.base_scan, name)
        except AttributeError as e:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            ) from e


class ExtraDimsScanMixin(ABC):
    """A mixin for scans that work with more than just the "points" dimension. For
    example, the
    :class:`~vbeam.scan.optimized.apodization_filtered_scan.ApodizationFilteredScan`
    may filter out different points for each transmit, so it has an extra "transmits"
    dimension.

    This is just to give vbeam enough information to automatically create beamformers
    from more advanced scan setups."""

    @property
    @abstractmethod
    def flattened_points_dimensions(self) -> Sequence[str]:
        """The named dimensions of the array returned by
        ``Scan.get_points(flatten=True)``"""

    @property
    @abstractmethod
    def required_dimensions_for_unflatten(self) -> Sequence[str]:
        "The named dimensions required as axis-arguments by :meth:`Scan.unflatten`."

    @property
    def dimensions_after_unflatten(self) -> Sequence[str]:
        "The named dimensions of the array returned by :meth:`Scan.unflatten`."
        return ["width", "height"]



from vbeam.apodization.combine import combine_apodizations
from vbeam.apodization.mla import MLAApodization
from vbeam.apodization.no_apodization import NoApodization
from vbeam.apodization.plane_wave import (
    PlaneWaveReceiveApodization,
    PlaneWaveTransmitApodization,
)
from vbeam.apodization.rtb import RTBApodization
from vbeam.apodization.tx_rx_apodization import TxRxApodization
from vbeam.apodization.util import get_apodization_values
from vbeam.apodization.window import (
    Bartlett,
    Hamming,
    Hanning,
    NoWindow,
    Rectangular,
    Tukey,
    Tukey25,
    Tukey50,
    Tukey75,
    Tukey80,
    Window,
)

__all__ = [
    "combine_apodizations",
    "MLAApodization",
    "NoApodization",
    "PlaneWaveReceiveApodization",
    "PlaneWaveTransmitApodization",
    "RTBApodization",
    "TxRxApodization",
    "get_apodization_values",
    "Bartlett",
    "Hamming",
    "Hanning",
    "NoWindow",
    "Rectangular",
    "Tukey",
    "Tukey25",
    "Tukey50",
    "Tukey75",
    "Tukey80",
    "Window",
]



import operator
from typing import Sequence, Set, Union

from spekk.transformations import Transformation

from vbeam.data_importers.setup import SignalForPointSetup
from vbeam.fastmath import numpy as np
from vbeam.scan import CoordinateSystem
from vbeam.scan import util as scan_util
from vbeam.scan.advanced import ExtraDimsScanMixin
from vbeam.util.transformations import *
from vbeam.util.vmap import apply_binary_operation_across_axes


def do_nothing(f):
    return f


def vectorize_over_datacube(
    setup: SignalForPointSetup,
    *,  # Always use keyword arguments for remaining args for clarity/readability
    ignore: Union[Sequence[str], Set[str]] = (),
) -> Transformation:
    """Return a :class:`Transformation` that vectorizes it to run for all points,
    receivers, transmits, and frames (if part of the spec).

    It assumes that the spec follows the convention of naming the dimensions "points",
    "receivers", "transmits", and "frames". A check is made whether the spec has the
    dimension before vectorizing over it — if the spec does not have the dimension then
    it is ignored."""
    ignore = set(ignore)
    return compose(
        *[
            ForAll(dim)
            for dim in ["points", "receivers", "senders", "transmits", "frames"]
            if setup.spec.has_dimension(dim) and dim not in ignore
        ]
    )


def sum_over_dimensions(
    setup: SignalForPointSetup,
    *,  # Always use keyword arguments for remaining args for clarity/readability
    keep: Union[Sequence[str], Set[str]] = (),
) -> Transformation:
    """Return a :class:`Transformation` that makes the function sum over receivers and
    transmits, unless they should be kept (as specified by :param:`keep`).

    Any dimension in :param:`keep` is not summed over. Dimensions that are not in the
    spec are also ignored.

    Dimensions needed to recombine the points using the points optimizer are always
    kept."""
    keep = set(keep)
    if isinstance(setup.scan, ExtraDimsScanMixin):
        keep |= set(setup.scan.required_dimensions_for_unflatten)
    return Apply(
        np.sum,
        [
            Axis(dim)
            for dim in ["receivers", "senders", "transmits"]
            if dim not in keep and setup.spec.has_dimension(dim)
        ],
    )


def unflatten_points(setup: SignalForPointSetup) -> Transformation:
    """Return a :class:`Transformation` that makes the function convert the flattened
    points back to the original scan shape with "width" and "height" dimensions
    (assuming a 2D scan).

    Automatically handles point optimizers."""
    if isinstance(setup.scan, ExtraDimsScanMixin):
        return Apply(
            setup.scan.unflatten,
            *[Axis(dim) for dim in setup.scan.required_dimensions_for_unflatten[:-1]],
            Axis(
                setup.scan.required_dimensions_for_unflatten[-1],
                becomes=setup.scan.dimensions_after_unflatten,
            ),
        )
    else:
        return Apply(setup.scan.unflatten, Axis("points", becomes=["width", "height"]))


def compensate_for_apodization_overlap(setup: SignalForPointSetup) -> Transformation:
    """Return a :class:`Transformation` that makes the function divide its result by
    the RTB apodization overlap for each pixel."""
    if isinstance(setup.scan, ExtraDimsScanMixin):
        apodization_overlap = setup.get_apodization_values(
            setup.scan.required_dimensions_for_unflatten
        )
        apodization_overlap = setup.scan.unflatten(
            apodization_overlap,
            *range(len(setup.scan.required_dimensions_for_unflatten)),
        )
    else:
        apodization_overlap = setup.get_apodization_values(["points"])
        apodization_overlap = setup.scan.unflatten(apodization_overlap)

    return Apply(
        lambda result, width_axis, height_axis: apply_binary_operation_across_axes(
            result, apodization_overlap, operator.truediv, [width_axis, height_axis]
        ),
        Axis("width", keep=True),
        Axis("height", keep=True),
    )


def scan_convert(setup: SignalForPointSetup) -> Transformation:
    """Return a :class:`Transformation` that makes the function apply scan conversion
    to its result if the scan is defined in polar coordinates."""
    if setup.scan.coordinate_system != CoordinateSystem.POLAR:
        return do_nothing
    return Apply(
        scan_util.scan_convert,
        setup.scan,
        Axis("width", keep=True),
        Axis("height", keep=True),
    )



import vbeam.scan.advanced as advanced
from vbeam.scan.advanced import ApodizationFilteredScan, ScanConvertedSectorScan
from vbeam.scan.base import CoordinateSystem, Scan
from vbeam.scan.linear_scan import LinearScan, linear_scan
from vbeam.scan.sector_scan import SectorScan, sector_scan
from vbeam.scan.util import scan_convert

__all__ = [
    "advanced",
    "ApodizationFilteredScan",
    "ScanConvertedSectorScan",
    "CoordinateSystem",
    "Scan",
    "LinearScan",
    "linear_scan",
    "SectorScan",
    "sector_scan",
    "scan_convert",
]



from typing import Callable, Iterable, Optional

from vbeam.core import Apodization
from vbeam.fastmath import numpy as np
from vbeam.fastmath.traceable import traceable_dataclass


@traceable_dataclass(("apodizations",), ("combiner",))
class CombinedApodization(Apodization):
    apodizations: Iterable[Apodization]
    combiner: Optional[Callable[[np.ndarray], float]] = None

    def __call__(self, *args, **kwargs) -> float:
        """Multiply the result of calling the Apodization objects."""
        values = np.array([apod(*args, **kwargs) for apod in self.apodizations])
        if self.combiner is not None:
            return self.combiner(values)
        return np.prod(values)


def combine_apodizations(
    *apodizations: Iterable[Apodization],
    combiner: Optional[Callable[[np.ndarray], float]] = None,
) -> CombinedApodization:
    """Return a new Apodization object that combines all the given apodizations.
    
    By default, apodizations are combined by taking the product of all their results."""
    return CombinedApodization(apodizations, combiner)
